import os
from os.path import join, exists, isfile, isdir
import shutil
from urllib import request
from urllib.error import URLError
from utils import AttributeZipFile

from make_config import TOOLCHAIN_CONFIG
from utils import merge_directory

def download_toolchain(directory):
	os.makedirs(directory, exist_ok=True)
	archive = join(directory, "toolchain.zip")

	if not exists(archive):
		url = "https://codeload.github.com/zheka2304/innercore-mod-toolchain/zip/deploy"
		print("Downloading Inner Core Mod Toolchain: " + url)
		try:
			request.urlretrieve(url, archive)
		except URLError:
			from task import error
			error("Check your network connection!", 1)
		except BaseException as err:
			print(err)
			error("Inner Core Mod Toolchain installation not completed due to above error.", 2)
	else:
		print("'toolchain.zip' already exists in temporary directory.")

def might_be_updated(directory = None):
	commit_path = TOOLCHAIN_CONFIG.get_path("toolchain/bin/.commit")
	if not isfile(commit_path):
		return True
	if directory is not None and isfile(join(directory, "toolchain.zip")):
		return True
	try:
		with open(join(commit_path)) as commit_file:
			response = request.urlopen("https://raw.githubusercontent.com/zheka2304/innercore-mod-toolchain/deploy/toolchain/toolchain/bin/.commit")
			return not perform_diff(response.read().decode("utf-8"), commit_file.read())
	except URLError:
		return False
	except BaseException as err:
		print(err)
		return True

def perform_diff(a, b):
	return str(a).strip() == str(b).strip()

def extract_toolchain(directory):
	archive = join(directory, "toolchain.zip")
	with AttributeZipFile(archive, "r") as zip_ref:
		zip_ref.extractall(directory)

	branch = join(directory, "innercore-mod-toolchain-deploy")
	if not exists(branch):
		print("Inner Core Mod Toolchain extracted 'innercore-mod-toolchain-deploy' folder not found.")
		from task import error
		error("Retry operation or extract 'toolchain.zip' manually.", 3)
	toolchain = TOOLCHAIN_CONFIG.get_path("..")

	if exists(join(branch, ".github")):
		merge_directory(join(branch, ".github"), join(toolchain, ".github"))
	for filename in os.listdir(branch):
		above = join(branch, filename)
		if not isfile(above):
			continue
		merge_directory(above, join(toolchain, filename))
	accept_squash_and_replace = TOOLCHAIN_CONFIG.get_value("updateAcceptReplaceConfiguration", True)
	merge_directory(join(branch, "toolchain"), join(toolchain, "toolchain"), accept_squash_and_replace, ["toolchain", "toolchain.json"], True, accept_squash_and_replace)
	merge_directory(join(branch, "toolchain/toolchain"), join(toolchain, "toolchain/toolchain"))
	if isdir(join(toolchain, "toolchain-sample-mod")) and isdir(join(branch, "toolchain-sample-mod")):
		shutil.rmtree(join(toolchain, "toolchain-sample-mod"))
		shutil.move(join(branch, "toolchain-sample-mod"), join(toolchain, "toolchain-sample-mod"))

	shutil.rmtree(branch, ignore_errors=True)
	os.remove(archive)

def update_toolchain():
	directory = TOOLCHAIN_CONFIG.get_path("toolchain/temp")
	if might_be_updated(directory):
		commit_path = TOOLCHAIN_CONFIG.get_path("toolchain/bin/.commit")
		download_toolchain(directory)
		commit = None
		if isfile(commit_path):
			with open(commit_path) as file:
				commit = file.read().strip()
		extract_toolchain(directory)
		if not isfile(commit_path):
			print("Successfully installed! But corresponding 'toolchain/bin/.commit' not found, futher update will be installed without any prompt.")
		else:
			with open(commit_path) as file:
				branch_commit = file.read().strip()
			if commit is not None:
				print(f"Successfully installed {branch_commit[:7]} above {commit[:7]} revision!")
			else:
				print(f"Successfully installed under {branch_commit[:7]} revision!")


if __name__ == "__main__":
	update_toolchain()



ESC = "\u001B["
OSC = "\u001B]"

CURSOR_LEFT = ESC + "G"
CURSOR_SAVE_POSITION = ESC + "s"
CURSOR_RESTORE_POSITION = ESC + "u"
CURSOR_GET_POSITION = ESC + "6n"
CURSOR_NEXT_LINE = ESC + "E"
CURSOR_PREV_LINE = ESC + "F"
CURSOR_HIDE = ESC + "?25l"
CURSOR_SHOW = ESC + "?25h"

ERASE_END_LINE = ESC + "K"
ERASE_START_LINE = ESC + "1K"
ERASE_LINE = ESC + "2K"
ERASE_DOWN = ESC + "J"
ERASE_UP = ESC + "1J"
ERASE_SCREEN = ESC + "2J"
SCROLL_UP = ESC + "S"
SCROLL_DOWN = ESC + "T"

CLEAR_SCREEN = "\u001Bc"

from platform import system
CLEAR_TERMINAL = f"{ERASE_SCREEN}{ESC}0f" if system() == "Windows" else f"{ERASE_SCREEN}{ESC}3J{ESC}H"

BEEP = "\u0007"

def cursor_to(x, y = None):
	if x is None:
		raise TypeError("The `x` argument is required")
	if y is None:
		return ESC + str(x + 1) + "G"
	return ESC + str(y + 1) + ";" + str(x + 1) + "H"

def cursor_move(x, y = 0):
	if x is None:
		raise TypeError("The `x` argument is required")
	returnValue = ""
	if x < 0:
		returnValue += ESC + str(-x) + "D"
	elif x > 0:
		returnValue += ESC + str(x) + "C"
	if y < 0:
		returnValue += ESC + str(-y) + "A"
	elif y > 0:
		returnValue += ESC + str(y) + "B"
	return returnValue

def cursor_up(count = 1):
	return ESC + str(count) + "A"

def cursor_down(count = 1):
	return ESC + str(count) + "B"

def cursor_forward(count = 1):
	return ESC + str(count) + "C"

def cursor_backward(count = 1):
	return ESC + str(count) + "D"

def erase_lines(count):
	clear = ""
	for i in range(count - 1):
		clear += ERASE_LINE + (cursor_up() if i < count - 1 else "")
	if count > 0:
		clear += CURSOR_LEFT
	return clear

def link(text, url = None):
	return f"{OSC}8;;{url if url is not None else text}{BEEP}{text}{OSC}8;;{BEEP}"

def image(base64, options = {}):
	returnValue = OSC + "1337;File=inline=1"
	if "width" in options:
		returnValue += ";width=" + str(options["width"])
	if "height" in options:
		returnValue += ";height=" + str(options["height"])
	if "preserveAspectRatio" in options and options["preserveAspectRatio"] == False:
		returnValue += ";preserveAspectRatio=0"
	return returnValue + ":" + base64 + BEEP



import os
from os.path import join, relpath, basename
import platform
import re
import socket
import subprocess
from glob import glob

from make_config import MAKE_CONFIG, TOOLCHAIN_CONFIG
from hash_storage import output_storage
from shell import Progress, Shell, select_prompt
from ansi_escapes import link

def get_modpack_push_directory():
	directory = MAKE_CONFIG.get_value("pushTo", accept_prototype=False)
	if directory is None:
		directory = TOOLCHAIN_CONFIG.get_value("pushTo")
		if directory is not None:
			directory = join(directory, "mods", basename(MAKE_CONFIG.current_project))
	if directory is None:
		TOOLCHAIN_CONFIG.set_value("pushTo", setup_modpack_directory())
		TOOLCHAIN_CONFIG.save()
		if MAKE_CONFIG.get_value("pushTo") is None:
			from task import error
			error("Nothing may be selected in modpack, nothing to do.")
		return get_modpack_push_directory()
	if "/horizon/packs/" not in directory and not MAKE_CONFIG.get_value("adb.pushAnyLocation", False):
		print(
			f"Push directory {directory} looks suspicious, it does not belong to Horizon packs directory. " +
			"This action may easily corrupt all content inside, allow it only if you know what are you doing."
		)
		which = select_prompt(
			"What will you do?",
			"Choice another modpack",
			"Push it anyway",
			"No questions, always push",
			"Nothing", fallback=3
		)
		if which == 0:
			TOOLCHAIN_CONFIG.remove_value("pushTo")
			MAKE_CONFIG.remove_value("pushTo")
			return get_modpack_push_directory()
		elif which == 2:
			TOOLCHAIN_CONFIG.set_value("adb.pushAnyLocation", True)
			TOOLCHAIN_CONFIG.save()
			print("This may be changed in your toolchain.json config.")
		elif which == 3:
			print("Pushing aborted.")
			return None
	return directory

def ls_pack(path):
	list = [path + "/innercore"] if "mods" in ls(path + "/innercore") else []
	return list + [path + "/modpacks/" + directory for directory in ls(path + "/modpacks")]

def person_readable_modpack_name(path):
	what = path.split("/")[::-1]
	try:
		suffix = " (internal)" if "Android/data" in path else ""
		if what[0] == "com.mojang":
			return "Legacy Core Engine" + suffix
		if what[0] == "innercore":
			return what[1] + suffix
		if what[3] == "packs":
			return f"{basename(path)} in {what[2]}" + suffix
		return "/".join(what[0:2][::-1]) + suffix
	except IndexError:
		pass
	return basename(path)

def setup_modpack_directory(directories = []):
	pack_directories = ls("/storage/emulated/0/games/horizon/packs")
	for directory in pack_directories:
		directories += ls_pack("/storage/emulated/0/games/horizon/packs/" + directory)
	pack_directories = ls("/storage/emulated/0/Android/data/com.zheka.horizon/files/horizon/packs")
	for directory in pack_directories:
		directories += ls_pack("/storage/emulated/0/Android/data/com.zheka.horizon/files/horizon/packs/" + directory)
	if "mods" in ls("/storage/emulated/0/games/com.mojang"):
		directories.append("/storage/emulated/0/games/com.mojang")
	if len(directories) == 0:
		print(
			"It seems your device doesn't contain any Inner Core pack directory. " +
			"If not, describe output path manually in toolchain.json."
		)
		return None
	which = select_prompt("Which modpack will be used?", *[
		person_readable_modpack_name(directory) for directory in directories
	])
	return None if which is None else directories[which]

def ls(path, *args):
	try:
		pipe = subprocess.run(adb_command + [
			"shell", "ls", path
		] + list(args), text=True, check=True, capture_output=True)
	except subprocess.CalledProcessError as err:
		if err.returncode != 1:
			print("adb shell ls failed with code", err.returncode)
		return []
	return pipe.stdout.rstrip().splitlines()

def push(directory, push_unchanged = False):
	shell = Shell()
	progress = Progress("Pushing")
	shell.interactables.append(progress)
	items = [relpath(path, directory) for path in glob(directory + "/*") if push_unchanged or output_storage.is_path_changed(path)]
	if len(items) == 0:
		progress.seek(1, "Nothing to push")
		shell.enter()
		shell.leave()
		return 0

	dst_root = get_modpack_push_directory()
	if dst_root is None:
		return 1

	shell.enter()
	dst_root = dst_root.replace("\\", "/")
	if not dst_root.startswith("/"):
		dst_root = "/" + dst_root
	src_root = directory.replace("\\", "/")

	percent = 0
	from task import devnull
	for filename in items:
		src = src_root + "/" + filename
		dst = dst_root + "/" + filename
		progress.seek(percent / len(items), "Pushing " + filename)
		shell.render()
		subprocess.call(adb_command + [
			"shell", "rm", "-r", dst
		], stderr=devnull, stdout=devnull)
		result = subprocess.call(adb_command + [
			"push", src, dst
		], stderr=devnull, stdout=devnull)
		percent += 1

		if result != 0:
			progress.seek(0.5, f"Failed {filename} with code {result}")
			shell.render()
			shell.leave()
			return result

	if not push_unchanged:
		output_storage.save()
	progress.seek(1, "Pushed")
	shell.render()
	shell.leave()
	return result

def make_locks(*locks):
	dst = get_modpack_push_directory()
	if dst is None:
		return -1

	for lock in locks:
		lock = join(dst, lock).replace("\\", "/")
		result = subprocess.call(adb_command + [
			"shell", "touch", lock
		])
		if result != 0:
			return result
	return 0

def ensure_server_running(retry = 0):
	try:
		from task import devnull
		subprocess.run([
			TOOLCHAIN_CONFIG.get_adb(),
			"start-server"
		], check=True, stdout=devnull, stderr=devnull)
		return True
	except subprocess.CalledProcessError as err:
		if retry >= 3:
			print("adb start-server failed with code", err.returncode)
			return False
		return ensure_server_running(retry + 1)

STATE_UNKNOWN = -1
STATE_DEVICE_CONNECTED = 0
STATE_NO_DEVICES = 1
STATE_DISCONNECTED = 2
STATE_DEVICE_AUTHORIZING = 3

def which_state(what = None):
	if what is None:
		return STATE_UNKNOWN
	try:
		return {
			"no devices": STATE_NO_DEVICES,
			"device": STATE_DEVICE_CONNECTED,
			"authorizing": STATE_DEVICE_AUTHORIZING
		}[what]
	except KeyError: # offline
		return STATE_DISCONNECTED

def get_device_state():
	try:
		pipe = subprocess.run([
			TOOLCHAIN_CONFIG.get_adb(),
			"get-state"
		], text=True, check=True, capture_output=True)
	except subprocess.CalledProcessError as err:
		if err.returncode == 1:
			return STATE_NO_DEVICES
		print("adb get-state failed with code", err.returncode)
		return STATE_UNKNOWN
	return which_state(pipe.stdout.strip())

def get_device_serial():
	try:
		pipe = subprocess.run([
			TOOLCHAIN_CONFIG.get_adb(),
			"get-serialno"
		], text=True, check=True, capture_output=True)
	except subprocess.CalledProcessError as err:
		print("adb get-serialno failed with code", err.returncode)
		return None
	return pipe.stdout.strip()

def device_list():
	try:
		pipe = subprocess.run([
			TOOLCHAIN_CONFIG.get_adb(),
			"devices", "-l"
		], text=True, check=True, capture_output=True)
	except subprocess.CalledProcessError as err:
		print("adb devices failed with code", err.returncode)
		return None
	data = pipe.stdout.rstrip().splitlines()
	data.pop(0)
	devices = []
	for device in data:
		device = re.split(r"\s+", device)
		devices.append({
			"serial": device[0],
			"state": which_state(device[1]),
			"data": device[2:]
		})
	return devices

def person_readable_device_name(device):
	if "data" in device:
		for property in device["data"]:
			what = property.partition(":")
			if what[0] == "model":
				return f"{device['serial']} ({what[2]})"
	return device["serial"]

def which_device_will_be_connected(*devices, state_not_matter = False):
	devices = [device for device in devices
		if device["state"] == STATE_DEVICE_CONNECTED
		or device["state"] == STATE_DEVICE_AUTHORIZING
		or state_not_matter]
	if len(devices) < 2:
		return None if len(devices) == 0 else devices[0]
	which = select_prompt("Which device will be used?", *[
		person_readable_device_name(device) for device in devices
	] + ["I don't see my device"])
	return None if which is None or which == len(devices) else devices[which]

def get_ip():
	make = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
	make.settimeout(0)
	try:
		make.connect(("10.253.254.255", 1))
		ip = make.getsockname()[0]
	except Exception:
		ip = "127.0.0.1"
	finally:
		make.close()
	return ip

def get_adb_command():
	ensure_server_running()
	devices = TOOLCHAIN_CONFIG.get_value("devices", [])
	from task import devnull
	if len(devices) > 0:
		subprocess.run([
			TOOLCHAIN_CONFIG.get_adb(),
			"disconnect"
		], stdout=devnull, stderr=devnull)
	for device in devices:
		if isinstance(device, dict):
			try:
				subprocess.run([
					TOOLCHAIN_CONFIG.get_adb(),
					"connect",
					f"{device['ip']}:{device['port']}" if "port" in device else device["ip"]
				], timeout=3.0, stdout=devnull, stderr=devnull)
			except subprocess.TimeoutExpired:
				print("Timeout")
	list = device_list()
	if list is not None:
		itwillbe = []
		for device in list:
			if device["serial"] in devices:
				itwillbe.append(device)
		device = which_device_will_be_connected(*(list if len(itwillbe) == 0 else itwillbe))
		if device is not None:
			return get_adb_command_by_serial(device["serial"])
	if MAKE_CONFIG.get_value("adb.doNothingIfDisconnected", False):
		from task import error
		error("Not found connected devices, nothing to do.", 0)
	which = setup_externally(True) if len(devices) > 0 else setup_device_connection()
	if which is None:
		from task import error
		error("Nothing will happened, adb set up interrupted.", 1)
	return which

def get_adb_command_by_serial(serial):
	ensure_server_running()
	devices = TOOLCHAIN_CONFIG.get_value("devices", [])
	if not serial in devices:
		try:
			from ipaddress import ip_address
			ip_address(serial.partition(":")[0])
		except ValueError:
			devices.append(serial)
			TOOLCHAIN_CONFIG.set_value("devices", devices)
			TOOLCHAIN_CONFIG.save()
	return [
		TOOLCHAIN_CONFIG.get_adb(),
		"-s", serial
	]

def get_adb_command_by_tcp(ip, port = None, skip_error = False):
	ensure_server_running()
	if get_adb_command_by_serialno_type("-e") is None:
		if skip_error:
			return None
		try:
			if input("Are you sure want to save it? [N/y] ")[:1].lower() != "y":
				return print("Abort.")
		except KeyboardInterrupt:
			return print("Abort.")
	device = {
		"ip": ip
	}
	if port is not None:
		device["port"] = port
	devices = TOOLCHAIN_CONFIG.get_value("devices", [])
	if not device in devices:
		devices.append(device)
		TOOLCHAIN_CONFIG.set_value("devices", devices)
		TOOLCHAIN_CONFIG.save()
	return [
		TOOLCHAIN_CONFIG.get_adb(),
		"-e"
	]

def get_adb_command_by_serialno_type(which):
	serial = subprocess.run([
		TOOLCHAIN_CONFIG.get_adb(),
		which, "get-serialno"
	], text=True, capture_output=True)
	if serial.returncode != 0:
		print("adb get-serialno failed with code", serial.returncode)
		return None
	return get_adb_command_by_serial(serial.stdout.rstrip())

def setup_device_connection():
	not_connected_any_device = len(TOOLCHAIN_CONFIG.get_value("devices", [])) == 0
	if not_connected_any_device:
		print(
			"Howdy! " +
			"Before starting we're must set up your devices, don't you think so? " +
			"Let's configure some connections."
		)
	which = select_prompt("How connection will be performed?", *[
		"I've connected device via cable",
		"Over air/network will be best",
		"Everything already performed"
	] + (["Wha.. I don't understand!"] if not_connected_any_device else []) + [
		"It will be performed later"
	])
	return setup_via_usb() if which == 0 else \
		setup_via_network() if which == 1 else \
		setup_externally() if which == 2 else \
		setup_how_to_use() if which == 3 and \
			not_connected_any_device else None

def setup_via_usb():
	try:
		print("Listening device via cable...")
		print(f"* Press Ctrl+{'Z' if platform.system() == 'Windows' else 'C'} to leave")
		from task import devnull
		subprocess.run([
			TOOLCHAIN_CONFIG.get_adb(),
			"wait-for-usb-device"
		], check=True, timeout=90.0, stdout=devnull, stderr=devnull)
		command = get_adb_command_by_serialno_type("-d")
		if command is not None:
			return command
	except subprocess.CalledProcessError as err:
		print("adb wait-for-usb-device failed with code", err.returncode)
	except subprocess.TimeoutExpired:
		print("Timeout")
	except KeyboardInterrupt:
		print()
	return setup_device_connection()

def setup_via_network():
	which = select_prompt(
		"Which network type must be used?",
		"Just TCP by IP and PORT",
		"Ping network automatically",
		"Connect with pairing code",
		"Turn back", fallback=2
	)
	return setup_device_connection() if which == 3 else \
		setup_via_ping_localhost() if which == 1 else \
		setup_via_tcp_network(with_pairing_code=which == 2)

def setup_via_ping_localhost():
	ip = get_ip().rpartition(".")
	if len(ip[2]) == 0:
		print("Not availabled right now.")
		return setup_via_network()
	shell = Shell()
	progress = Progress(text="Connecting")
	shell.interactables.append(progress)
	shell.enter()
	accepted = []
	try:
		import asyncio
		asyncio.run(ping_async(ip, shell, progress, accepted))
	except ImportError:
		for index in range(256):
			if str(index) == ip[2]:
				continue
			next_ip = "{}.{}".format(ip[0], index)
			if ping_via_shell(next_ip, shell, progress, index):
				accepted.append(next_ip)
	shell.leave()
	from task import devnull
	subprocess.run([
		TOOLCHAIN_CONFIG.get_adb(),
		"disconnect"
	], stdout=devnull, stderr=devnull)
	latest = None
	for next in accepted:
		try:
			subprocess.run([
				TOOLCHAIN_CONFIG.get_adb(),
				"connect", next
			], check=True, timeout=5.0, stdout=devnull, stderr=devnull)
			command = get_adb_command_by_tcp(next, skip_error=True)
			if command is not None:
				latest = command
		except subprocess.CalledProcessError as err:
			print("adb connect failed with code", err.errorcode)
		except subprocess.TimeoutExpired:
			print("Timeout")
		except KeyboardInterrupt:
			print()
	if latest is not None:
		return latest
	if len(accepted) > 0:
		print("Found several variants to connection, default port not worked otherwise.")
		print("\n".join(accepted))
		print("You can try reconnect to one of it with specified port.")
		try:
			input()
		except KeyboardInterrupt:
			pass
	return setup_via_network()

def ping_via_shell(ip, shell, progress, index):
	progress.text = "Pinging " + ip
	progress.progress = index / 255
	shell.render()

	from task import devnull
	return subprocess.call([
		"ping",
		"-n" if platform.system() == "Windows" else "-c", "1",
		ip
	], stdout=devnull, stderr=devnull) == 0

async def ping(ip, shell, progress, index, accepted):
	progress.text = "Pinging " + ip
	progress.progress = index / 255
	shell.render()

	from task import devnull
	import asyncio
	coroutine = await asyncio.create_subprocess_shell(
		"ping " + ("-n" if platform.system() == "Windows" else "-c") + " 1 " + ip, stdout=devnull, stderr=devnull
	)
	await coroutine.wait()

	if coroutine.returncode == 0:
		accepted.append(ip)

async def ping_async(ip, shell, progress, accepted):
	import asyncio
	tasks = []
	for index in range(256):
		if str(index) == ip[2]:
			continue
		next_ip = "{}.{}".format(ip[0], index)
		task = asyncio.ensure_future(ping(next_ip, shell, progress, index, accepted))
		tasks.append(task)
	await asyncio.gather(*tasks, return_exceptions=True)

def setup_via_tcp_network(ip = None, port = None, pairing_code = None, with_pairing_code = False):
	if ip is None:
		print("You are connected via", get_ip())
		try:
			tcp = input("Specify address: IP[:PORT] ")
			if len(tcp) == 0:
				return setup_via_network()
		except KeyboardInterrupt:
			return setup_via_network()
		tcp = tcp.split(":")
		ip = tcp[0]
		port = tcp[1] if len(tcp) > 1 else port
	if with_pairing_code or pairing_code is not None:
		if pairing_code is None:
			try:
				pairing_code = input("Specify pairing code: ")
				if len(tcp) == 0:
					return setup_via_tcp_network(ip, port)
			except KeyboardInterrupt:
				return setup_via_network()
		try:
			subprocess.run([
				TOOLCHAIN_CONFIG.get_adb(),
				"pair",
				f"{ip}:{port}" if port is not None else ip,
				pairing_code
			], check=True, stderr=devnull, stdout=devnull)
		except subprocess.CalledProcessError as err:
			print("adb pair failed with code", err.errorcode)
		except KeyboardInterrupt:
			print()
	from task import devnull
	subprocess.run([
		TOOLCHAIN_CONFIG.get_adb(),
		"disconnect"
	], stdout=devnull, stderr=devnull)
	try:
		subprocess.run([
			TOOLCHAIN_CONFIG.get_adb(),
			"connect",
			f"{ip}:{port}" if port is not None else ip
		], check=True, timeout=10.0, stdout=devnull, stderr=devnull)
		command = get_adb_command_by_tcp(ip, port)
		return command if command is not None else setup_via_tcp_network()
	except subprocess.CalledProcessError as err:
		print("adb connect failed with code", err.errorcode)
	except subprocess.TimeoutExpired:
		print("Timeout")
	except KeyboardInterrupt:
		print()
	return setup_via_network()

def setup_externally(skip_input = False):
	state = get_device_state()
	if state == STATE_DEVICE_CONNECTED or state == STATE_DEVICE_AUTHORIZING:
		serial = get_device_serial()
		if serial is not None:
			if not serial in TOOLCHAIN_CONFIG.get_value("devices", []):
				return get_adb_command_by_serial(serial)
			else:
				print("Connected device already saved, maybe another available too.")
	else:
		print("Not found connected devices, resolving everything...")
	list = device_list()
	if list is None:
		return setup_device_connection()
	device = which_device_will_be_connected(*list, state_not_matter=True)
	if device is None:
		print("Nope, nothing to perform here.")
		if not skip_input:
			try:
				input()
			except KeyboardInterrupt:
				pass
		return setup_device_connection()
	return get_adb_command_by_serial(device["serial"])

def setup_how_to_use():
	print(
		"Android Debug Bridge (adb) is a versatile command-line tool that lets you communicate with a device. " +
		"The adb command facilitates a variety of device actions, such as installing and debugging apps, " +
		"and it provides access to a Unix shell that you can use to run a variety of commands on a device."
	)
	print(link("https://developer.android.com/studio/command-line/adb"))
	try:
		input()
	except KeyboardInterrupt:
		pass
	return setup_device_connection()

adb_command = get_adb_command()



import os
from os.path import join, exists, abspath, isfile, isdir
import shutil
import sys
from zipfile import ZipFile, ZipInfo

def ensure_directory(directory):
	if isfile(directory):
		os.remove(directory)
	if not exists(directory):
		os.makedirs(directory, exist_ok=True)

def ensure_file_dir(file):
	ensure_directory(abspath(join(file, "..")))

def clear_directory(directory):
	ensure_directory(directory)
	shutil.rmtree(directory)

def copy_file(src, dst):
	ensure_file_dir(dst)
	try:
		shutil.copy(src, dst)
	except shutil.SameFileError:
		pass

def move_file(src, dst):
	ensure_file_dir(dst)
	shutil.move(src, dst)

def copy_directory(path, destination, clear_dst = False, replacement = True, ignore_list = [], relative_path = None):
	ensure_directory(destination)
	if clear_dst:
		clear_directory(destination)
	if not exists(destination):
		os.makedirs(destination, exist_ok=True)
	for filename in os.listdir(path):
		relative_file = join(relative_path, filename) if relative_path is not None else filename
		if filename in ignore_list or relative_file in ignore_list:
			continue
		input = join(path, filename)
		if input in ignore_list:
			continue
		output = join(destination, filename)
		if isfile(input) and exists(output) and not replacement:
			continue
		if isdir(input):
			if isfile(output):
				os.remove(output)
			copy_directory(input, output, clear_dst, replacement, ignore_list, relative_file)
		elif not isdir(output):
			try:
				shutil.copy2(input, output)
			except shutil.SameFileError:
				pass

def merge_directory(src, dst, accept_squash = True, ignore_list = [], only_parent_ignore = False, accept_replace = True):
	if ((isdir(src) and isfile(dst)) or (isdir(dst) and isfile(src))) and accept_squash:
		if isfile(dst + ".bak"):
			os.remove(dst + ".bak")
		elif isdir(dst + ".bak"):
			shutil.rmtree(dst + ".bak")
		shutil.move(dst, dst + ".bak")
	if not isdir(src):
		if not exists(dst):
			shutil.move(src, dst)
		return
	for filename in os.listdir(src):
		if filename in ignore_list:
			continue
		above = join(src, filename)
		behind = join(dst, filename)
		if isfile(above) and (not exists(behind) or (isfile(behind) and accept_replace)):
			if exists(behind):
				os.remove(behind)
			shutil.move(above, behind)
		elif isdir(above) or accept_squash:
			merge_directory(above, behind, accept_squash, ignore_list if not only_parent_ignore else [], only_parent_ignore, accept_replace)

def get_all_files(directory, extensions = ()):
	all_files = []
	for dirpath, dirnames, filenames in os.walk(directory):
		for filename in filenames:
			if len(extensions) == 0:
				all_files.append(abspath(join(dirpath, filename)))
			else:
				for extension in extensions:
					if len(filename) >= len(extension) and filename[-len(extension):] == extension:
						all_files.append(abspath(join(dirpath, filename)))
						break
	return all_files

def ensure_not_whitespace(what, fallback = None):
	return fallback if what is None or len(what) == 0 or what.isspace() else what

def name_to_identifier(name, delimiter = ""):
	previous_char_lower = False
	previous_chars_upper = 0
	identifier = ""
	for char in name:
		if char.isalpha() or char.isdecimal():
			if (char.isupper() and previous_char_lower) or (char.islower() and previous_chars_upper > 1):
				identifier += delimiter
			identifier += char.lower()
		elif len(delimiter) > 0 and not identifier.endswith(delimiter):
			identifier += delimiter
		previous_char_lower = char.islower()
		previous_chars_upper = previous_chars_upper + 1 if char.isupper() else 0
	return identifier

def get_project_folder_by_name(directory, name):
	folder = name_to_identifier(name, "-").strip("-")
	return get_next_filename(directory, folder, "-") if len(folder) > 0 else None

def get_next_filename(directory, name, delimiter = ""):
	buffer_name = name
	buffer_index = 0
	while exists(join(directory, buffer_name)):
		buffer_index += 1
		buffer_name = f"{name}{delimiter}{buffer_index}"
	return buffer_name

def shortcodes(str):
	from datetime import datetime
	date = datetime.now()
	str = str.replace("{datestamp}", date.strftime("%Y%m%d"))
	str = str.replace("{timestamp}", date.strftime("%H%M"))
	return str

def request_typescript():
	if shutil.which("tsc") is not None:
		return "typescript"
	try:
		if input("Do you want to enable TypeScript and ES6+ support (requires Node.js to build project)? [Y/n] ")[:1].lower() == "n":
			return "javascript"
	except KeyboardInterrupt:
		pass
	print("Updating TypeScript globally via npm...")
	os.system("npm install -g typescript")
	return request_typescript()

class AttributeZipFile(ZipFile):
	if sys.version_info < (3, 6):
		def extract(self, member, path = None, pwd = None):
			if not isinstance(member, ZipInfo):
				member = self.getinfo(member)

			if path is None:
				path = os.getcwd()

			targetpath = self._extract_member(member, path, pwd)

			attr = member.external_attr >> 16
			os.chmod(targetpath, attr)
			return targetpath

	else:
		def _extract_member(self, member, targetpath, pwd):
			if not isinstance(member, ZipInfo):
				member = self.getinfo(member)

			targetpath = super()._extract_member(member, targetpath, pwd)

			attr = member.external_attr >> 16
			if attr != 0:
				os.chmod(targetpath, attr)
			return targetpath



import sys
import os
from os.path import join, exists, basename, isfile, isdir, relpath
import platform
import time

from utils import ensure_directory, ensure_file_dir, clear_directory, copy_file, copy_directory
from make_config import MAKE_CONFIG, TOOLCHAIN_CONFIG

registered_tasks = {}
locked_tasks = {}
devnull = open(os.devnull, "w")

def lock_task(name, silent = True):
	path = TOOLCHAIN_CONFIG.get_path(f"toolchain/temp/lock/{name}.lock")
	ensure_file_dir(path)
	await_message = False

	if exists(path):
		while True:
			try:
				if exists(path):
					os.remove(path)
				break
			except IOError:
				if not await_message:
					await_message = True
					if not silent:
						sys.stdout.write(f"* Task {name} is locked by another process, waiting for it to unlock.")
					if name in locked_tasks:
						error("ERROR: Dead lock detected", code=-2)
				if not silent:
					sys.stdout.write(".")
					sys.stdout.flush()
				time.sleep(0.25)
	if await_message:
		if not silent:
			print("")
	open(path, "tw").close()
	locked_tasks[name] = open(path, "a")

def unlock_task(name):
	if name in locked_tasks:
		locked_tasks[name].close()
		del locked_tasks[name]
	path = TOOLCHAIN_CONFIG.get_path(f"toolchain/temp/lock/{name}.lock")
	if isfile(path):
		os.remove(path)

def unlock_all_tasks():
	for name in list(locked_tasks.keys()):
		unlock_task(name)

def task(name, lock = None):
	if lock is None:
		lock = []

	def decorator(func):
		def caller(*args, **kwargs):
			lock_task(name, silent=False)
			for lock_name in lock:
				lock_task(lock_name, silent=False)
			if platform.system() == "Windows":
				os.system("color")
			print(f"\x1b[92m> Executing task: {name}\x1b[0m")
			task_result = func(*args, **kwargs)
			unlock_task(name)
			for lock_name in lock:
				unlock_task(lock_name)
			return task_result

		registered_tasks[name] = caller
		return caller

	return decorator

@task("compileNativeDebug", lock=["native", "cleanup", "push"])
def task_compile_native_debug(args = None):
	abi = MAKE_CONFIG.get_value("debugAbi", None)
	if abi is None:
		abi = "armeabi-v7a"
		print(f"* No 'debugAbi' value in toolchain.json config, using {abi} as default.")
	from native.native_build import compile_all_using_make_config
	return compile_all_using_make_config([abi])

@task("compileNativeRelease", lock=["native", "cleanup", "push"])
def task_compile_native_release(args = None):
	abis = MAKE_CONFIG.get_value("abis", [])
	if abis is None or not isinstance(abis, list) or len(abis) == 0:
		error(f"No 'abis' value in toolchain.json config, nothing will happened.")
	from native.native_build import compile_all_using_make_config
	return compile_all_using_make_config(abis)

@task("compileJavaDebug", lock=["java", "cleanup", "push"])
def task_compile_java_debug(args = None):
	from java.java_build import compile_all_using_make_config
	return compile_all_using_make_config(debug_build=True)

@task("compileJavaRelease", lock=["java", "cleanup", "push"])
def task_compile_java_release(args = None):
	from java.java_build import compile_all_using_make_config
	return compile_all_using_make_config(debug_build=False)

@task("buildScriptsDebug", lock=["script", "cleanup", "push"])
def task_build_scripts_debug(args = None):
	from script_build import build_all_scripts
	return build_all_scripts(debug_build=True)

@task("buildScriptsRelease", lock=["script", "cleanup", "push"])
def task_build_scripts_release(args = None):
	from script_build import build_all_scripts
	from hash_storage import output_storage, build_storage
	output_storage.last_hashes = {}
	build_storage.last_hashes = {}
	return build_all_scripts(debug_build=False)

@task("buildResources", lock=["resource", "cleanup", "push"])
def task_resources(args = None):
	from script_build import build_all_resources
	return build_all_resources()

@task("buildInfo", lock=["cleanup", "push"])
def task_build_info(args = None):
	import json
	from utils import shortcodes
	with open(MAKE_CONFIG.get_path("output/mod.info"), "w") as info_file:
		info = dict(MAKE_CONFIG.get_value("info", fallback={}))

		if "name" in info:
			info["name"] = shortcodes(info["name"])
		if "version" in info:
			info["version"] = shortcodes(info["version"])
		if "description" in info:
			info["description"] = shortcodes(info["description"])
		if "icon" in info:
			del info["icon"]
		if "instantLaunch" in info:
			info["instantLaunch"] = info["instantLaunch"]

		info_file.write(json.dumps(info, indent="\t") + "\n")
	icon_path = MAKE_CONFIG.get_value("info.icon")
	if icon_path is not None:
		icon_path = MAKE_CONFIG.get_absolute_path(icon_path)
		if isfile(icon_path):
			copy_file(icon_path, MAKE_CONFIG.get_path("output/mod_icon.png"))
		else:
			print("Icon in make.json", icon_path, "not found!")
	return 0

@task("buildAdditional", lock=["cleanup", "push"])
def task_build_additional(args = None):
	overall_result = 0
	for additional_dir in MAKE_CONFIG.get_value("additional", fallback=[]):
		if "source" in additional_dir and "targetDir" in additional_dir:
			for additional_path in MAKE_CONFIG.get_paths(additional_dir["source"]):
				if not exists(additional_path):
					print("Non existing additional path: " + additional_path)
					overall_result = 1
					break
				target = MAKE_CONFIG.get_path(join(
					"output",
					additional_dir["targetDir"],
					basename(additional_path)
				))
				if isdir(additional_path):
					copy_directory(additional_path, target)
				else:
					ensure_file_dir(target)
					copy_file(additional_path, target)
	return overall_result

@task("pushEverything", lock=["push"])
def task_push_everything(args = None):
	from device import push
	return push(MAKE_CONFIG.get_path("output"), MAKE_CONFIG.get_value("adb.pushUnchangedFiles", True))

@task("clearOutput", lock=["assemble", "push", "native", "java"])
def task_clear_output(args = None):
	clear_directory(MAKE_CONFIG.get_path("output"))
	return 0

@task("excludeDirectories", lock=["push", "assemble", "native", "java"])
def task_exclude_directories(args = None):
	for path in MAKE_CONFIG.get_value("excludeFromRelease", []):
		for exclude in MAKE_CONFIG.get_paths(join("output", path)):
			if isdir(exclude):
				clear_directory(exclude)
			elif isfile(exclude):
				os.remove(exclude)
	return 0

@task("buildPackage", lock=["push", "assemble", "native", "java"])
def task_build_package(args = None):
	import shutil
	output_dir = MAKE_CONFIG.get_path("output")
	name = basename(MAKE_CONFIG.current_project) if MAKE_CONFIG.current_project is not None else "unknown"
	output_dir_root_tmp = MAKE_CONFIG.get_build_path("package")
	output_dir_tmp = join(output_dir_root_tmp, name)
	ensure_directory(output_dir)
	if exists(output_dir_tmp):
		shutil.rmtree(output_dir_tmp)
	output_file_tmp = join(output_dir_root_tmp, "package.zip")
	ensure_file_dir(output_file_tmp)
	output_file = MAKE_CONFIG.get_path(name + ".icmod")
	if isfile(output_file):
		os.remove(output_file)
	if isfile(output_file_tmp):
		os.remove(output_file_tmp)
	shutil.copytree(output_dir, output_dir_tmp)
	shutil.make_archive(output_file_tmp[:-4], "zip", output_dir_root_tmp, name)
	shutil.rmtree(output_dir_tmp)
	os.rename(output_file_tmp, output_file)
	return 0

@task("launchHorizon")
def task_launch_horizon(args = None):
	from subprocess import call
	from device import adb_command
	call(adb_command + [
		"shell", "touch",
		"/storage/emulated/0/games/horizon/.flag_auto_launch"
	], stdout=devnull, stderr=devnull)
	return call(adb_command + [
		"shell", "monkey",
		"-p", "com.zheka.horizon",
		"-c", "android.intent.category.LAUNCHER", "1"
	], stdout=devnull, stderr=devnull)

@task("stopHorizon")
def stop_horizon(args = None):
	from subprocess import call
	from device import adb_command
	return call(adb_command + [
		"shell",
		"am",
		"force-stop",
		"com.zheka.horizon"
	], stdout=devnull, stderr=devnull)

@task("loadDocs")
def task_load_docs(args = None):
	from urllib.request import urlopen
	print("Downloading core-engine.d.ts")
	response = urlopen("https://docs.mineprogramming.org/headers/core-engine.d.ts")
	content = response.read().decode("utf-8")

	declaration_path = TOOLCHAIN_CONFIG.get_path("toolchain/declarations")
	if not exists(declaration_path):
		os.mkdir(declaration_path)
	with open(join(declaration_path, "core-engine.d.ts"), "w") as docs:
		docs.write(content)

	print("Complete!")
	return 0

@task("updateIncludes")
def task_update_includes(args = None):
	from script_build import build_all_make_scripts
	return build_all_make_scripts(only_tsconfig_rebuild=True, debug_build=True)

@task("configureADB")
def task_configure_adb(args = None):
	import device
	device.setup_device_connection()
	return 0

@task("newProject")
def task_new_project(args = None):
	from package import new_project

	index = new_project(MAKE_CONFIG.get_value("defaultTemplate", "../toolchain-mod"))
	if index is None:
		print()
		print("Abort.")
		exit(0)
	print("Successfully completed!")

	from project_manager import PROJECT_MANAGER
	try:
		if input("Select this project? [Y/n] ")[:1].lower() == "n":
			return 0
	except KeyboardInterrupt:
		pass
	PROJECT_MANAGER.select_project(index=index)
	return 0

@task("importProject")
def task_import_project(args = None):
	module = __import__("import")
	path = module.import_project(args[0] if args is not None and len(args) > 0 else None, args[1] if args is not None and len(args) > 1 else None)
	print("Project successfully imported!")

	from project_manager import PROJECT_MANAGER
	try:
		if input("Select this project? [Y/n] ")[:1].lower() == "n":
			return 0
	except KeyboardInterrupt:
		pass
	PROJECT_MANAGER.select_project(folder=relpath(path, TOOLCHAIN_CONFIG.root_dir))
	return 0

@task("removeProject", lock=["cleanup"])
def task_remove_project(args = None):
	from project_manager import PROJECT_MANAGER
	if PROJECT_MANAGER.how_much() == 0:
		error("Not found any project to remove.")

	print("Selected project will be deleted forever, please think twice before removing anything!")
	who = PROJECT_MANAGER.require_selection("Which project will be deleted?", "Do you really want to delete {}?", "I don't want it anymore")
	if who is None:
		exit(0)

	if PROJECT_MANAGER.how_much() > 1:
		try:
			if input("Do you really want to delete it? [Y/n] ")[:1].lower() == "n":
				print("Abort.")
				return 0
		except KeyboardInterrupt:
			pass

	try:
		location = TOOLCHAIN_CONFIG.get_absolute_path(who)
		PROJECT_MANAGER.remove_project(folder=who)
		from make_config import ToolchainMakeConfig
		from package import cleanup_relative_directory
		cleanup_relative_directory("toolchain/build/" + ToolchainMakeConfig.unique_folder_name(location))
	except ValueError:
		error(f"Folder '{who}' not found!")

	print("Project permanently deleted.")
	return 0

@task("selectProject", lock=["cleanup"])
def task_select_project(args = None):
	from project_manager import PROJECT_MANAGER
	if args is not None and len(args) > 0 and len(args[0]) > 0:
		if exists(args[0]):
			where = relpath(args[0], TOOLCHAIN_CONFIG.root_dir)
			if where == ".":
				error("Requested project path must be reference to mod, not toolchain itself.")
			if not exists(join(args[0], "make.json")):
				error("Not found make.json in requested folder, it not belongs to project yet.")
			PROJECT_MANAGER.select_project_folder(folder=where)
			return 0
		else:
			error("Requested project path does not exists.")

	if PROJECT_MANAGER.how_much() == 0:
		error("Not found any project to choice.")

	who = PROJECT_MANAGER.require_selection("Which project do you choice?", "Do you want to select {}?")
	if who is None:
		exit(0)

	try:
		PROJECT_MANAGER.select_project(folder=who)
	except ValueError:
		error(f"Folder '{who}' not found!")
	return 0

@task("updateToolchain")
def task_update_toolchain(args = None):
	from update import update_toolchain
	update_toolchain()
	from component import fetch_components, install_components
	upgradable = fetch_components()
	if len(upgradable) > 0:
		print("Found new updates for components: ", ", ".join(upgradable), ".", sep="")
		try:
			if input("Do you want to upgrade it? [Y/n] ")[:1].lower() == "n":
				print("Abort.")
				return 0
		except KeyboardInterrupt:
			pass
		install_components(upgradable)
	return 0

@task("componentIntegrity")
def task_component_integrity(args = None):
	from component import foreign
	return foreign()

@task("cleanup")
def task_cleanup(args = None):
	from package import cleanup_relative_directory
	if MAKE_CONFIG.current_project is not None:
		try:
			if input("Do you want to clear only selected project (everything cache will be cleaned otherwise)? [Y/n] ")[:1].lower() == "n":
				cleanup_relative_directory("toolchain/build")
				return 0
		except KeyboardInterrupt:
			pass
		cleanup_relative_directory("toolchain/build/" + MAKE_CONFIG.project_unique_name)
		cleanup_relative_directory("output", True)
		return 0
	else:
		try:
			if input("Do you want to clear all projects cache? [Y/n] ")[:1].lower() == "n":
				print("Abort.")
				return 0
		except KeyboardInterrupt:
			pass
	cleanup_relative_directory("toolchain/build")
	return 0

def error(message, code = -1):
	print(message, file=sys.stderr)
	unlock_all_tasks()
	print(message)
	exit(code)


if __name__ == "__main__":
	if "--help" in sys.argv:
		print("Usage: task.py <tasks> @ [arguments]")
		print("Executes declared by @task annotation required tasks.")
		exit(0)
	argv = sys.argv[1:]
	if "@" in argv:
		where = argv.index("@")
		args = argv[where + 1:]
		argv = argv[:where]
	else:
		args = None
	if len(argv) > 0:
		for task_name in argv:
			if task_name in registered_tasks:
				try:
					result = registered_tasks[task_name](args)
					if result != 0:
						print()
						error(f"* Task {task_name} failed with result {result}.", code=result)
				except BaseException as err:
					if isinstance(err, SystemExit):
						raise err

					import traceback
					traceback.print_exc()
					error(f"* Task {task_name} failed with above error!")
			else:
				print(f"* No such task: {task_name}.")
	else:
		error("* No tasks to execute.")
	unlock_all_tasks()



import json
import os
from os.path import exists, isdir, join, basename, relpath
import time

from utils import clear_directory, copy_directory, copy_file, ensure_not_whitespace, get_all_files, get_project_folder_by_name, name_to_identifier
from base_config import BaseConfig
from make_config import MAKE_CONFIG, TOOLCHAIN_CONFIG
from shell import Input, Interrupt, Notice, Progress, SelectiveShell, Entry, Separator, Shell, Switch, select_prompt
from project_manager import PROJECT_MANAGER

def get_path_set(pathes, error_sensitive = False):
	directories = []
	for path in pathes:
		for directory in MAKE_CONFIG.get_paths(path):
			if isdir(directory):
				directories.append(directory)
			else:
				if error_sensitive:
					print(f"Declared invalid directory {path}, task will be terminated")
					return None
				else:
					print(f"Declared invalid directory {path}, it will be skipped")
	return directories

def assemble_additional_directories():
	result = 0
	output_dir = MAKE_CONFIG.get_path("output")
	for additional_dir in MAKE_CONFIG.get_value("additional", []):
		if "sources" not in additional_dir or "pushTo" not in additional_dir:
			print("Invalid formatted additional directory json", additional_dir)
			result = -1
			break
		dst_dir = join(output_dir, additional_dir["pushTo"])
		clear_directory(dst_dir)
		source_directories = get_path_set(additional_dir["sources"], error_sensitive=True)
		if source_directories is None:
			print("Some additional directories are invalid, nothing will happened")
			result = -1
			break
		for source_dir in source_directories:
			copy_directory(source_dir, dst_dir)
	return result

def cleanup_relative_directory(path, project = False):
	start_time = time.time()
	clear_directory(MAKE_CONFIG.get_path(path) if project else TOOLCHAIN_CONFIG.get_path(path))
	print(f"Completed {basename(path)} cleanup in {int((time.time() - start_time) * 100) / 100}s")

def select_template():
	if len(PROJECT_MANAGER.templates) <= 1:
		if len(PROJECT_MANAGER.templates) == 0:
			print("Please, ensure that `projectLocations` property in your toolchain.json contains any folder with template.json.")
			from task import error
			error("Not found any templates, nothing to do.")
		return PROJECT_MANAGER.templates[0]
	return select_prompt(
		"Which template do you want?",
		*PROJECT_MANAGER.templates,
		fallback=0, what_not_which=True
	)

def new_project(template = "../toolchain-mod"):
	if template is None or not exists(TOOLCHAIN_CONFIG.get_absolute_path(template)):
		return new_project(template=select_template())
	template_make_path = TOOLCHAIN_CONFIG.get_absolute_path(template + "/template.json")
	try:
		with open(template_make_path) as template_make:
			template_config = BaseConfig(json.loads(template_make.read()))
	except BaseException as err:
		print(err)
		if len(PROJECT_MANAGER.templates) > 1:
			return new_project(None)
		from task import error
		error(f"Malformed {template}/template.json, nothing to do.")

	have_template = TOOLCHAIN_CONFIG.get_value("template") is not None
	always_skip_description = TOOLCHAIN_CONFIG.get_value("template.skipDescription", False)
	progress_step = 0.5 if have_template and always_skip_description else 0.33 if have_template or always_skip_description else 0.25
	print("Inner Core Mod Toolchain", end="")

	class NameObserver(Shell.Interactable):
		def __init__(self):
			Shell.Interactable.__init__(self, "name_observer")

		def observe_key(self, what):
			input = shell.get_interactable("name")
			self.directory = get_project_folder_by_name(TOOLCHAIN_CONFIG.root_dir, input.read())
			header = shell.get_interactable("header")
			header.size = (1 if self.directory is None else 0) + (0 if len(PROJECT_MANAGER.templates) > 1 else 1)
			location = shell.get_interactable("location")
			location.text = "" if self.directory is None else "It will be in " + self.directory + "\n"
			progress = shell.get_interactable("step")
			progress.progress = 0 if self.directory is None else progress_step
			progress.text = " " + "Name your creation".center(45) + (" " if self.directory is None else ">")
			shell.blocked_in_page = self.directory is None
			return False

	shell = SelectiveShell()
	shell.interactables.append(Notice("Create new project"))
	if len(PROJECT_MANAGER.templates) > 1:
		shell.interactables += [
			Separator("header"),
			Entry("template", "Choose template")
		]
	else:
		shell.interactables.append(Separator("header", size=2))
	shell.interactables += [
		Input("name", "Name: ", TOOLCHAIN_CONFIG.get_value("template.name", ""), template=template_config.get_value("info.name")),
		Notice("location"),
		NameObserver(),
		Progress("step")
	]
	if not always_skip_description:
		shell.interactables += [
			Input("author", "Author: ", TOOLCHAIN_CONFIG.get_value(
				"template.author", template_config.get_value("info.author", "")
			)),
			Input("version", "Version: ", TOOLCHAIN_CONFIG.get_value(
				"template.version", template_config.get_value("info.version", "1.0")
			)),
			Input("description", "Description: ", TOOLCHAIN_CONFIG.get_value(
				"template.description", template_config.get_value("info.description", "")
			)),
			Switch("client_side", "Client side only", TOOLCHAIN_CONFIG.get_value(
				"template.clientOnly", template_config.get_value("info.clientOnly", False)
			)),
			Separator(),
			Progress(progress=progress_step * 2, text="<" + "Configure details".center(45) + (">" if not have_template else "+"))
		]
	if not have_template:
		shell.interactables += [
			Notice("You can override template by setting up `template`"),
			Notice("property in your toolchain.json, it will be automatically"),
			Notice("applied when new project is being created."),
			Notice("Properties are still same make.json `info` property."),
			Separator(),
			Progress(progress=progress_step * (3 if not always_skip_description else 2), text="<" + "Friendly advice".center(45) + "+")
		]

	shell.interactables.append(Interrupt())
	observer = shell.get_interactable("name_observer")
	observer.observe_key(None)
	try:
		shell.loop()
	except KeyboardInterrupt:
		shell.leave()
		return None
	if shell.what() == "template":
		return new_project(None)
	if not hasattr(observer, "directory") or observer.directory is None:
		from task import error
		error("Not found `directory` property in observer!")
	print(f"Copying template '{template}' to '{observer.directory}'")
	return PROJECT_MANAGER.create_project(
		template, observer.directory,
		shell.get_interactable("name").read(),
		shell.get_interactable("author").read(),
		shell.get_interactable("version").read(),
		shell.get_interactable("description").read(),
		shell.get_interactable("client_side").checked
	)

def resolve_make_format_map(make_obj, path):
	make_obj_info = make_obj["info"] if "info" in make_obj else {}
	identifier = name_to_identifier(basename(path))
	while len(identifier) > 0 and identifier[0].isdecimal():
		identifier = identifier[1:]
	package_prefix = name_to_identifier(make_obj_info["author"]) if "author" in make_obj_info else "icmods"
	while len(package_prefix) > 0 and package_prefix[0].isdecimal():
		package_prefix = package_prefix[1:]
	package_suffix = name_to_identifier(make_obj_info["name"]) if "name" in make_obj_info else identifier
	while len(package_suffix) > 0 and package_suffix[0].isdecimal():
		package_suffix = package_suffix[1:]
	return {
		"identifier": ensure_not_whitespace(identifier, "whoami"),
		"packageSuffix": ensure_not_whitespace(package_suffix, "mod"),
		"packagePrefix": package_prefix,
		**make_obj_info,
		"clientOnly": "true" if "clientOnly" in make_obj_info and make_obj_info["clientOnly"] else "false"
	}

def setup_project(make_obj, template, path):
	makemap = resolve_make_format_map(make_obj, path)
	dirmap = { template: "" }
	for dirpath, dirnames, filenames in os.walk(template):
		for dirname in dirnames:
			dir = join(dirpath, dirname)
			dirmap[dir] = relpath(dir, template)
			try:
				dirmap[dir] = dirmap[dir].format_map(makemap)
			except BaseException:
				print(f"Source {dirmap[dir]} contains malformed name!")
			os.mkdir(join(path, dirmap[dir]))
		for filename in filenames:
			if dirpath == template and filename == "template.json":
				continue
			file = join(path, join(dirmap[dirpath], filename))
			copy_file(join(dirpath, filename), file)
	for source in get_all_files(path, extensions=(".json", ".js", ".ts", "manifest", ".java", ".cpp")):
		with open(source, "r") as source_file:
			lines = source_file.readlines()
		for index in range(len(lines)):
			try:
				lines[index] = lines[index].format_map(makemap)
			except BaseException:
				pass
		with open(source, "w") as source_file:
			source_file.writelines(lines)

def select_project(variants, prompt = "Which project do you want?", selected = None):
	if prompt is not None:
		print(prompt, end="")
	shell = SelectiveShell(infinite_scroll=True)
	for variant in variants:
		shell.interactables.append(Entry(variant, variant if selected != variant else f"\x1b[7m{variant}\x1b[0m"))
	try:
		shell.loop()
	except KeyboardInterrupt:
		shell.leave()
		print()
		return None
	try:
		print((prompt + " " if prompt is not None else "") + "\x1b[2m" + shell.what() + "\x1b[0m")
	except ValueError:
		pass
	return shell.what()



from os.path import exists, splitext, basename, isfile, isdir, join
from functools import cmp_to_key

from utils import clear_directory, copy_file, copy_directory, request_typescript
from make_config import MAKE_CONFIG, TOOLCHAIN_CONFIG
from mod_structure import mod_structure
from includes import Includes

def build_source(source_path, target_path, language, includes_file, debug_build = False):
	includes = Includes.invalidate(source_path, includes_file, debug_build)
	return includes.build(target_path, language)

def build_all_scripts(debug_build = False):
	mod_structure.cleanup_build_target("script_source")
	mod_structure.cleanup_build_target("script_library")

	allowed_languages = []
	if len(MAKE_CONFIG.get_filtered_list("sources", "language", ("typescript"))) > 0 or MAKE_CONFIG.get_value("denyJavaScript", True):
		if request_typescript() == "typescript":
			allowed_languages.append("typescript")
		if not exists(TOOLCHAIN_CONFIG.get_path("toolchain/declarations")):
			print("\x1b[93mNot found toolchain/declarations, in most cases build will be failed, please install it via tasks.\x1b[0m")
	if not MAKE_CONFIG.get_value("denyJavaScript", True):
		allowed_languages.append("javascript")

	if len(allowed_languages) == 0:
		from task import error
		error("TypeScript is required by default, if you want to build legacy JavaScript, change `denyJavaScript` property in your make.json or toolchain.json config.")
	return build_all_make_scripts(allowed_languages=allowed_languages, debug_build=debug_build)

def libraries_first(a, b):
	la = a["type"] == "library"
	lb = b["type"] == "library"
	return 0 if la == lb else -1 if la else 1

def build_all_make_scripts(only_tsconfig_rebuild = False, allowed_languages = ["typescript"], debug_build = False):
	overall_result = 0
	sources = MAKE_CONFIG.get_value("sources", fallback=[])
	sources = sorted(sources, key=cmp_to_key(libraries_first))

	for item in sources:
		_source = item["source"]
		_target = item["target"] if "target" in item else None
		_type = item["type"]
		_includes = item["includes"] if "includes" in item else ".includes"

		if _type not in ("main", "launcher", "preloader", "instant", "custom", "library"):
			print(f"Skipped invalid source with type {_type}")
			overall_result = 1
			continue

		for source_path in MAKE_CONFIG.get_paths(_source):
			if not exists(source_path):
				print(f"Skipped non existing source {_source}")
				overall_result = 1
				continue
			if not (isdir(source_path) or source_path.endswith(".js") or source_path.endswith(".ts")):
				continue

			target_type = "script_library" if _type == "library" else "script_source"
			if _target is None:
				target_path = basename(source_path)
				if isfile(source_path):
					target_path = splitext(target_path)[0]
				target_path += ".js"
			else:
				target_path = _target

			# translate make.json source type to build.config source type
			declare = {
				"sourceType": "mod" if _type == "main" else _type
			}

			if "api" in item and _type != "preloader":
				declare["api"] = item["api"]
			if "optimizationLevel" in item:
				declare["optimizationLevel"] = min(max(int(item["optimizationLevel"]), -1), 9)
			if "sourceName" in item:
				declare["sourceName"] = item["sourceName"]

			try:
				dot_index = target_path.rindex(".")
				target_path = target_path[:dot_index] + "{}" + target_path[dot_index:]
			except ValueError:
				target_path += "{}"

			if not only_tsconfig_rebuild:
				destination_path = mod_structure.new_build_target(
					target_type,
					target_path,
					source_type=_type,
					declare=declare
				)
			mod_structure.update_build_config_list("compile")
			if not only_tsconfig_rebuild:
				if isfile(source_path):
					copy_file(source_path, destination_path)
				else:
					overall_result += build_source(
						source_path, destination_path,
						item["language"] if "language" in item and item["language"] in allowed_languages else allowed_languages[0],
						_includes, debug_build
					)
			elif isdir(source_path):
				from includes import temp_directory
				includes = Includes.invalidate(source_path, _includes, debug_build)
				includes.create_tsconfig(join(temp_directory, basename(target_path)))

	return overall_result

def build_all_resources():
	mod_structure.cleanup_build_target("resource_directory")
	mod_structure.cleanup_build_target("gui")
	mod_structure.cleanup_build_target("minecraft_resource_pack")
	mod_structure.cleanup_build_target("minecraft_behavior_pack")
	overall_result = 0

	for resource in MAKE_CONFIG.get_value("resources", fallback=[]):
		if "path" not in resource or "type" not in resource:
			print("Skipped invalid source json", resource)
			overall_result = 1
			continue
		for source_path in MAKE_CONFIG.get_paths(resource["path"]):
			if not exists(source_path):
				print("Skipped non existing resource", resource["path"])
				overall_result = 1
				continue
			resource_type = resource["type"]
			if resource_type not in ("resource_directory", "gui", "minecraft_resource_pack", "minecraft_behavior_pack"):
				print("Skipped invalid resource with type", resource_type)
				overall_result = 1
				continue
			resource_name = resource["target"] if "target" in resource else basename(source_path)
			resource_name += "{}"

			if resource_type in ("resource_directory", "gui"):
				target = mod_structure.new_build_target(
					resource_type,
					resource_name,
					declare={
						"resourceType": "resource" if resource_type == "resource_directory" else resource_type
					}
				)
			else:
				target = mod_structure.new_build_target(
					resource_type,
					resource_name,
					exclude=True,
					declare_default={
						"resourcePacksDir": mod_structure.get_target_directories("minecraft_resource_pack")[0],
						"behaviorPacksDir": mod_structure.get_target_directories("minecraft_behavior_pack")[0]
					}
				)
			clear_directory(target)
			copy_directory(source_path, target)
	mod_structure.update_build_config_list("resources")
	return overall_result


if __name__ == "__main__":
	build_all_resources()
	build_all_scripts(debug_build=True)



from glob import glob
import os
from os.path import exists, join, basename, isfile, isdir, relpath
import sys
import json

from utils import copy_directory, copy_file, ensure_directory, get_project_folder_by_name
from base_config import BaseConfig

def load_mod_info(make_obj, directory):
	mod_info = join(directory, "mod.info")
	mod_info_obj = {
		"name": basename(directory if directory != "." else os.getcwd())
	} if not "info" in make_obj else make_obj["info"]
	if not isfile(mod_info):
		make_obj["info"] = mod_info_obj
		return

	with open(mod_info, "r", encoding="utf-8") as mod_info_file:
		mod_info_prototype = json.loads(mod_info_file.read())
		if "name" in mod_info_prototype:
			mod_info_obj["name"] = mod_info_prototype["name"]
		if "version" in mod_info_prototype:
			mod_info_obj["version"] = mod_info_prototype["version"]
		if "author" in mod_info_prototype:
			mod_info_obj["author"] = mod_info_prototype["author"]
		if "description" in mod_info_prototype:
			mod_info_obj["description"] = mod_info_prototype["description"]
		if "instantLaunch" in mod_info_prototype:
			mod_info_obj["instantLaunch"] = mod_info_prototype["instantLaunch"]
	make_obj["info"] = mod_info_obj

def load_build_config(make_obj, source, destination):
	build_config = join(source, "build.config")
	destination_copy_tuples = []
	if not isfile(build_config):
		return destination_copy_tuples

	with open(build_config, "r", encoding="utf-8") as build_config_file:
		build_config_obj = json.loads(build_config_file.read())
		config = BaseConfig(build_config_obj)

		default_config_api = config.get_value("defaultConfig.api")
		if default_config_api is not None or not "api" in make_obj:
			make_obj["api"] = default_config_api if default_config_api is not None else "CoreEngine"
		default_config_optimization_level = config.get_value("defaultConfig.optimizationLevel")
		if default_config_optimization_level is not None:
			make_obj["optimizationLevel"] = default_config_optimization_level
		default_config_setup_script = config.get_value("defaultConfig.setupScript")
		if default_config_setup_script is not None:
			make_obj["setupScript"] = default_config_setup_script
			destination_copy_tuples.append((join(source, default_config_setup_script), join(destination, default_config_setup_script)))

		assets_dir = join(destination, "assets")
		if not "resources" in make_obj:
			make_obj["resources"] = []

		for directory in config.get_filtered_list("resources", "resourceType", ("resource", "gui")):
			if directory["resourceType"] == "resource":
				directory["resourceType"] = "resource_directory"
			path_stripped = directory["path"].strip("/")
			path = join(*path_stripped.split("/"))
			destination_copy_tuples.append((join(source, path), join(assets_dir, path)))
			make_obj["resources"].append({
				"path": "assets/" + path_stripped,
				"type": directory["resourceType"]
			})

		libs = config.get_value("defaultConfig.libraryDir", "lib").strip("/")
		libs_directory = join(source, *libs.split("/"))
		if isdir(libs_directory):
			destination_copy_tuples.append((libs_directory, join(destination, "lib")))
			if not "sources" in make_obj or len(make_obj["sources"]) == 0:
				make_obj["sources"] = [
					{
						"source": "lib/*",
						"type": "library",
						"language": "javascript"
					}
				]
		elif not "sources" in make_obj:
			make_obj["sources"] = []

		for directory in config.get_filtered_list("compile", "sourceType", ("mod", "launcher", "preloader", "instant", "custom", "library")):
			if directory["sourceType"] == "mod":
				directory["sourceType"] = "main"
			toolchain_source = {
				"type": directory["sourceType"],
				"language": "javascript"
			}
			if "api" in directory:
				toolchain_source["api"] = directory["api"]
			if "optimizationLevel" in directory:
				toolchain_source["optimizationLevel"] = directory["optimizationLevel"]
			if "sourceName" in directory:
				toolchain_source["sourceName"] = directory["sourceName"]
			path_stripped = directory["path"].split("/")
			build_dirs = config.get_filtered_list("buildDirs", "targetSource", (directory["path"]))
			if len(build_dirs) > 0:
				toolchain_source["source"] = build_dirs[0]["dir"].strip("/")
				build_path_stripped = toolchain_source["source"].split("/")
				toolchain_source["target"] = directory["path"]
				destination_copy_tuples.append((join(source, *build_path_stripped), join(destination, *build_path_stripped)))
			else:
				toolchain_source["source"] = directory["path"]
				destination_copy_tuples.append((join(source, *path_stripped), join(destination, *path_stripped)))
			make_obj["sources"].append(toolchain_source)

		if not "compile" in make_obj:
			make_obj["compile"] = []

		for directory in config.get_value("javaDirs", []):
			path_stripped = directory["path"].strip("/")
			path = join(source, *path_stripped.split("/"))
			destination_copy_tuples.append((path, join(destination, "java", basename(path))))
			make_obj["compile"].append({
				"source": "java/" + basename(path),
				"type": "java"
			})

		for directory in config.get_value("nativeDirs", []):
			path_stripped = directory["path"].strip("/")
			path = join(source, *path_stripped.split("/"))
			destination_copy_tuples.append((path, join(destination, "native", basename(path))))
			make_obj["compile"].append({
				"source": "native/" + basename(path),
				"type": "native"
			})

		if not "additional" in make_obj or len(make_obj["additional"]) == 0:
			make_obj["additional"] = [
				{
					"source": "assets/root/*",
					"targetDir": "."
				}
			]

	return destination_copy_tuples

def copy_tuple_directories(tuples, source, destination):
	ignore_list = [".dex", "make.json", "mod_icon.png", "mod.info", "build.config", "assets/root"]
	for path, output in tuples:
		if path == output:
			continue
		if isfile(path) and (not exists(output) or isfile(output)):
			if exists(output):
				print("Replacing", basename(output))
			copy_file(path, output)
		elif isdir(path) and (not exists(output) or isdir(output)):
			if exists(output):
				print("Merging", basename(output))
			copy_directory(path, output)
		else:
			print("Destination and input", basename(path), "pathes conflicts!")
		ignore_list.append(relpath(path, source))
	mod_icon = join(source, "mod_icon.png")
	if exists(mod_icon) and isfile(mod_icon):
		copy_file(mod_icon, join(destination, "mod_icon.png"))
	for file in glob(join(source, "*.md")):
		if exists(file) and isfile(file):
			relative_path = relpath(file, source)
			copy_file(file, join(destination, relative_path))
			ignore_list.append(relative_path)
	git = join(source, "LICENSE")
	if exists(git) and isfile(git):
		copy_file(git, join(destination, "LICENSE"))
		ignore_list.append("LICENSE")
	git = join(source, ".gitignore")
	if exists(git) and isfile(git):
		copy_file(git, join(destination, ".gitignore"))
	git = join(source, ".git")
	if exists(git) and isdir(git):
		copy_directory(git, join(destination, ".git"))
		ignore_list.append(".git")
	github = join(source, ".github")
	if exists(github) and isdir(github):
		copy_directory(github, join(destination, ".github"))
		ignore_list.append(".github")
	additional_path = join(destination, "assets", "root")
	copy_directory(source, additional_path, ignore_list=ignore_list)
	for dirpath, dirnames, filenames in os.walk(additional_path, False):
		for dirname in dirnames:
			directory = join(dirpath, dirname)
			if len(os.listdir(directory)) == 0:
				os.rmdir(directory)
	if len(os.listdir(additional_path)) == 0:
		os.removedirs(additional_path)

def merge_json(left, right):
	if not isinstance(right, dict):
		return right
	for key in right:
		if isinstance(left[key], list) and isinstance(right[key], list):
			left[key].extend(right[key])
			continue
		left[key] = merge_json(left[key], right[key])
	return right

def import_project(path = None, destination = None):
	if path is None:
		print("Specify absolute or relative path to toolchain folder that must be imported as project, it may be Inner Core mod or already exists Mod Toolchain folder.")
		try:
			path = input("Which directory will be imported? ")
		except KeyboardInterrupt:
			print("Abort.")
			exit(0)
	destination_may_changed = destination is None
	if destination_may_changed:
		from make_config import TOOLCHAIN_CONFIG
		toolchain = TOOLCHAIN_CONFIG.root_dir
		destination = join(toolchain, get_project_folder_by_name(toolchain, basename(path)))

	if exists(path) and isfile(path):
		path = join(path, "..")
	if not isdir(path):
		print("Input path not exists or specified!")
		exit(1)
	if exists(destination) and isfile(destination):
		print("Destination path is file!")
		exit(2)
	if not (isfile(join(path, "build.config")) or isfile(join(path, "make.json"))):
		print("Not found build.config or make.json entry to import, nothing to do!")
		exit(3)
	print(f"Importing '{path}' into {basename(destination)}")
	make_obj = {}

	if destination_may_changed:
		make = join(path, "make.json")
		if exists(make) and isfile(make):
			print("Reading make.json to resolve destination")
			with open(make, "r", encoding="utf-8") as make_file:
				make_obj = json.loads(make_file.read())

		print("Resolving mod.info to resolve destination")
		load_mod_info(make_obj, path)

		potential_name = make_obj["info"]["name"] if "info" in make_obj and "name" in make_obj["info"] else None
		if potential_name is not None:
			potential_name = get_project_folder_by_name(toolchain, potential_name)
			if potential_name != basename(destination):
				destination = join(toolchain, potential_name)
				print(f"Output directory changed to {potential_name}")

	make_project = join(destination, "make.json")
	if isfile(make_project):
		print("Reading output make.json")
		with open(make_project, "r", encoding="utf-8") as make_file:
			make_obj = json.loads(make_file.read())

	if not destination_may_changed or isfile(make_project):
		make = join(path, "make.json")
		if exists(make) and isfile(make):
			print("Merging with existing make.json")
			with open(make, "r", encoding="utf-8") as make_file:
				make_obj = merge_json(make_obj, json.loads(make_file.read()))

		print("Resolving mod.info")
		load_mod_info(make_obj, path)

	print("Resolving build.config")
	tuples = load_build_config(make_obj, path, destination)

	print("Flushing make.json")
	ensure_directory(destination)
	with open(make_project, "w", encoding="utf-8") as make_file:
		make_file.write(json.dumps(make_obj, indent="\t") + "\n")

	try:
		if destination == path and input("Do you want to copy reassigned directories in directory itself? [N/y] ")[:1].lower() != "y":
			print("Abort.")
			exit(0)
	except KeyboardInterrupt:
		print("Abort.")
		exit(0)

	print("Copying files and directories")
	copy_tuple_directories(tuples, path, destination)

	return destination


if __name__ == "__main__":
	if "--help" in sys.argv:
		print("Usage: import.py <path> [destination]")
		print("Performs conversion between mod.info, build.config and make.json,")
		print("merges directories if few configurations exists.")
		exit(0)

	import_project(sys.argv[1] if len(sys.argv) > 1 else None, sys.argv[2] if len(sys.argv) > 2 else None)
	print("Project successfully imported!")



import os
from os.path import isfile, isdir, join, dirname
from errno import ENOENT
import json

from hashlib import md5
from make_config import MAKE_CONFIG

class HashStorage:
	last_hashes = {}
	hashes = {}

	def __init__(self, file):
		self.file = file
		if isfile(file):
			with open(file, "r") as input:
				self.last_hashes = json.load(input)

	def get_path_hash(self, path):
		key = self.path_to_key(path)
		if key in self.hashes:
			return self.hashes[key]

		if isfile(path):
			hash = HashStorage.get_file_hash(path)
		elif isdir(path):
			hash = HashStorage.get_directory_hash(path)
		else:
			raise FileNotFoundError(ENOENT, os.strerror(ENOENT), path)

		self.hashes[key] = hash
		return hash

	@staticmethod
	def get_directory_hash(directory):
		total = md5()
		for dirpath, dirnames, filenames in os.walk(directory):
			for filename in filenames:
				filepath = join(dirpath, filename)
				total.update(open(filepath, "rb").read())
				"""
				with open(filepath, "rb") as f:
					for chunk in iter(lambda: f.read(4096), b""):
						total.update(chunk)
				"""
		return total.hexdigest()

	@staticmethod
	def get_file_hash(file):
		return md5(open(file, "rb").read()).hexdigest()

	def save(self):
		os.makedirs(dirname(self.file), exist_ok=True)
		with open(self.file, "w") as output:
			json.dump(self.hashes, output, indent="\t")

	def is_path_changed(self, path):
		key = self.path_to_key(path)
		hash = self.get_path_hash(path)
		return key not in self.last_hashes or self.last_hashes[key] != hash

	def path_to_key(self, path):
		return md5(path.encode("utf-8")).hexdigest()


build_storage = HashStorage(MAKE_CONFIG.get_build_path(".buildhashes"))
output_storage = HashStorage(MAKE_CONFIG.get_build_path(".outputhashes"))



import sys
try:
	import termios
	import tty
except ImportError:
	import msvcrt

from ansi_escapes import *

class Shell():
	offset = 0
	line = 0

	def __init__(self, stdin = sys.stdin, stdout = sys.stdout):
		self.stdin = stdin
		self.stdout = stdout
		self.eof_when_enter = False
		self.interactables = []

	def read(self, count = 1):
		return self.stdin.read(count)

	def readline(self, count = 1):
		return self.stdin.readline(count)

	def input_raw(self, count = 1):
		try:
			fd = self.stdin.fileno()
			term_attrs = termios.tcgetattr(fd)
		except NameError:
			pass
		try:
			try:
				tty.setraw(fd)
				key = self.stdin.read(count)
			except NameError:
				key = msvcrt.getwch()
				count -= 1
				while count > 0:
					key += msvcrt.getwch()
					count -= 1
		finally:
			try:
				termios.tcsetattr(fd, termios.TCSADRAIN, term_attrs)
			except NameError:
				pass
		return key

	def input(self, count = 1):
		key = self.input_raw(count)
		if key == "\x03" or key == "\x1a": # Ctrl+C or Ctrl+Z
			raise KeyboardInterrupt()
		return key

	def inputline(self, count = 1):
		buffer = ""
		while count > 0:
			key = self.input()
			if ord(key) in {10, 13}: # Enter
				if count > 1:
					buffer += "\n"
				count -= 1
			else:
				buffer += key
		return buffer

	def write(self, value):
		self.stdout.write(value)
		value = str(value)
		self.line += value.count("\n")
		try:
			where = value.rindex("\n") + 1
			self.offset = len(value[where:])
		except ValueError:
			self.offset += len(value)

	def up(self, count):
		self.write(cursor_up(count))

	def down(self, count):
		self.write(cursor_down(count))

	def right(self, count):
		self.write(cursor_forward(count))

	def left(self, count):
		self.write(cursor_backward(count))

	def clear(self):
		if self.line > 0:
			self.stdout.write(erase_lines(self.line + 1))
			self.line = 0
		self.offset = 0

	def get_interactable(self, criteria):
		try:
			return self.interactables[criteria]
		except IndexError:
			pass
		except TypeError:
			pass
		for interactable in self.interactables:
			if interactable.key == criteria:
				return interactable
		raise ValueError(criteria)

	def observe(self, key):
		if self.eof_when_enter and ord(key) in {10, 13}: # Enter
			raise EOFError()
		return False

	def draw(self, interactable):
		return interactable.render(self, self.offset, self.line)

	def touch(self, interactable, key):
		return interactable.observe_key(key)

	def render(self):
		self.clear()
		self.write("\n")
		for interactable in self.interactables:
			self.draw(interactable)

	def enter(self):
		self.hide_cursor()
		self.render()

	def loop(self):
		self.enter()
		while True:
			try:
				key = self.input(1)
				observed = False
				for interactable in self.interactables:
					observed = self.touch(interactable, key) or observed
				if not observed:
					observed = self.observe(key)
				if observed:
					self.render()
			except EOFError:
				break
			except KeyboardInterrupt as err:
				self.leave()
				raise err
		self.leave()

	def leave(self):
		self.show_cursor()

	def hide_cursor(self):
		self.write(CURSOR_HIDE)

	def show_cursor(self):
		self.write(CURSOR_SHOW)

	class Interactable():
		def __init__(self, key):
			self.key = key

		def observe_key(self, what):
			return False

		def render(self, shell, offset, line):
			pass

		def lines(self, shell):
			return 0

class InteractiveShell(Shell):
	page_buffer_offset = 0
	global_buffer_offset = 0
	page = 1

	def __init__(self, stdin = sys.stdin, stdout = sys.stdout, infinite_scroll = False, lines_per_page = 6):
		Shell.__init__(self, stdin, stdout)
		self.infinite_scroll = infinite_scroll
		self.lines_per_page = lines_per_page

	def observe(self, raw):
		observed = Shell.observe(self, raw)
		if raw != "\x1b" and raw != "\xe0":
			return observed
		key = self.input_raw(1)
		if raw == "\xe0":
			if key == "M": # Right
				self.turn_forward()
			elif key == "K": # Left
				self.turn_backward()
			else:
				return observed
			return True
		if key != "[":
			return observed
		joy = self.input_raw(1)
		if joy == "C": # Right
			self.turn_forward()
		elif joy == "D": # Left
			self.turn_backward()
		else:
			return observed
		return True

	def turn_forward(self):
		if self.global_buffer_offset + self.page_buffer_offset >= len(self.interactables):
			if self.infinite_scroll:
				self.global_buffer_offset = self.page_buffer_offset = 0
				self.page = 1
			return
		self.global_buffer_offset += self.page_buffer_offset
		self.page_buffer_offset = 0
		self.page += 1

	def turn_backward(self):
		index = self.global_buffer_offset
		if index == 0:
			if self.infinite_scroll:
				page = 1
				page_occupied_lines = 0
				page_buffer_offset = 0
				while index < len(self.interactables):
					lines = self.interactables[index].lines(self)
					if lines > self.lines_per_page or lines < 0:
						raise BufferError()
					if page_occupied_lines + lines > self.lines_per_page:
						page_occupied_lines = page_buffer_offset = 0
						page += 1
						continue
					page_occupied_lines += lines
					page_buffer_offset += 1
					index += 1
				self.global_buffer_offset = index - page_buffer_offset
				self.page_buffer_offset = 0
				self.page = page
			return
		page_occupied_lines = 0
		while index > 0:
			lines = self.interactables[index - 1].lines(self)
			if lines > self.lines_per_page or lines < 0:
				raise BufferError()
			if page_occupied_lines + lines > self.lines_per_page:
				break
			page_occupied_lines += lines
			index -= 1
		self.global_buffer_offset = index
		self.page_buffer_offset = 0
		self.page -= 1

	def draw(self, interactable, page, page_occupied_lines):
		if isinstance(interactable, InteractiveShell.Interactable):
			return interactable.render(self, self.offset, self.line, page, self.page_buffer_offset, page_occupied_lines)
		return Shell.draw(self, interactable)

	def render(self):
		self.clear()
		if len(self.interactables) == 0:
			return
		if self.global_buffer_offset >= len(self.interactables):
			raise IndexError()
		self.write("\n")
		page_occupied_lines = 0
		self.page_buffer_offset = 0
		while self.global_buffer_offset + self.page_buffer_offset < len(self.interactables):
			interactable = self.interactables[self.global_buffer_offset + self.page_buffer_offset]
			lines = interactable.lines(self)
			if lines > self.lines_per_page or lines < 0:
				raise BufferError()
			if page_occupied_lines + lines > self.lines_per_page:
				break
			self.draw(interactable, self.page, page_occupied_lines)
			page_occupied_lines += lines
			self.page_buffer_offset += 1

	def enter(self):
		self.global_buffer_offset = self.page_buffer_offset = 0
		self.page = 1
		Shell.enter(self)

	def leave(self):
		self.clear()
		Shell.leave(self)

	class Interactable(Shell.Interactable):
		def __init__(self, key):
			Shell.Interactable.__init__(self, key)

		def render(self, shell, offset, line, page = 0, index = -1, lines_before = -1):
			pass

class SelectiveShell(InteractiveShell):
	page_cursor_offset = -1
	pending_hover_offset = 0
	blocked_in_page = False

	def __init__(self, stdin = sys.stdin, stdout = sys.stdout, infinite_scroll = False, lines_per_page = 6):
		InteractiveShell.__init__(self, stdin, stdout, infinite_scroll, lines_per_page)
		self.eof_when_enter = True

	def turn_backward(self):
		if not self.blocked_in_page:
			InteractiveShell.turn_backward(self)
		self.page_cursor_offset += 1
		self.pending_hover_offset = -2

	def turn_forward(self):
		if not self.blocked_in_page:
			InteractiveShell.turn_forward(self)
		self.page_cursor_offset -= 1
		self.pending_hover_offset = 2

	def turn_up(self):
		if self.page_cursor_offset > 0 or self.infinite_scroll or self.global_buffer_offset > 0:
			self.pending_hover_offset = -1 if not self.blocked_in_page else -2

	def turn_down(self):
		if (self.page_cursor_offset < self.page_buffer_offset and self.which() < len(self.interactables) - 1) or self.infinite_scroll:
			self.pending_hover_offset = 1 if not self.blocked_in_page else 2

	def hover_previous(self):
		cursor_offset = min(self.page_cursor_offset, self.page_buffer_offset) - 1
		while cursor_offset >= 0:
			try:
				interactable = self.interactables[self.global_buffer_offset + cursor_offset]
			except IndexError:
				cursor_offset -= 1
				continue
			if isinstance(interactable, SelectiveShell.Selectable) and interactable.hoverable():
				self.page_cursor_offset = cursor_offset
				return True
			cursor_offset -= 1
		return False

	def hover_next(self):
		cursor_offset = max(self.page_cursor_offset, -1) + 1
		while cursor_offset < self.page_buffer_offset:
			try:
				interactable = self.interactables[self.global_buffer_offset + cursor_offset]
			except IndexError:
				cursor_offset += 1
				continue
			if isinstance(interactable, SelectiveShell.Selectable) and interactable.hoverable():
				self.page_cursor_offset = cursor_offset
				return True
			cursor_offset += 1
		return False

	def touch(self, interactable, key):
		if isinstance(interactable, SelectiveShell.Selectable):
			try:
				return interactable.observe_key(key, self.interactables.index(interactable) == self.which())
			except ValueError:
				pass
		return InteractiveShell.touch(self, interactable, key)

	def render(self):
		InteractiveShell.render(self)
		if self.pending_hover_offset != 0:
			if self.pending_hover_offset <= -1:
				if not self.hover_previous():
					if self.pending_hover_offset == -1:
						if self.infinite_scroll or self.global_buffer_offset > 0:
							self.page_cursor_offset = self.lines_per_page
							InteractiveShell.turn_backward(self)
							InteractiveShell.render(self)
							if not self.hover_previous():
								self.page_cursor_offset = -1
					elif not self.hover_next() and not self.hovered():
						self.page_cursor_offset = -1
				InteractiveShell.render(self)
			elif self.pending_hover_offset >= 1:
				if not self.hover_next():
					if self.pending_hover_offset == 1:
						self.page_cursor_offset = -1
						InteractiveShell.turn_forward(self)
						InteractiveShell.render(self)
						if not self.hover_next():
							self.page_cursor_offset = -1
					elif not self.hover_previous() and not self.hovered():
						self.page_cursor_offset = -1
				InteractiveShell.render(self)
			self.pending_hover_offset = 0

	def observe(self, raw):
		if self.eof_when_enter and ord(raw) in {10, 13}: # Enter
			if self.which() == -1:
				return False
			raise EOFError()
		observed = Shell.observe(self, raw)
		if raw != "\x1b" and raw != "\xe0":
			return observed
		key = self.input_raw(1)
		if raw == "\xe0":
			if key == "H": # Up
				self.turn_up()
			elif key == "P": # Down
				self.turn_down()
			elif key == "M": # Right
				self.turn_forward()
			elif key == "K": # Left
				self.turn_backward()
			else:
				return observed
			return True
		if key != "[":
			return observed
		joy = self.input_raw(1)
		if joy == "A": # Up
			self.turn_up()
		elif joy == "B": # Down
			self.turn_down()
		elif joy == "C": # Right
			self.turn_forward()
		elif joy == "D": # Left
			self.turn_backward()
		else:
			return observed
		return True

	def draw(self, interactable, page, page_occupied_lines):
		if isinstance(interactable, SelectiveShell.Selectable):
			return interactable.render(self, self.offset, self.line, page, self.page_buffer_offset, page_occupied_lines, self.page_cursor_offset == self.page_buffer_offset)
		return InteractiveShell.draw(self, interactable, page, page_occupied_lines)

	def enter(self):
		self.page_cursor_offset = -1
		self.pending_hover_offset = 2
		InteractiveShell.enter(self)

	def hovered(self):
		interactable = self.get_interactable(self.which())
		return interactable.hoverable() if isinstance(interactable, SelectiveShell.Selectable) else False

	def which(self):
		return self.global_buffer_offset + self.page_cursor_offset if self.page_cursor_offset != -1 else -1

	def what(self):
		try:
			return self.interactables[self.which()].key
		except IndexError:
			return None

	class Selectable(InteractiveShell.Interactable):
		def __init__(self, key):
			InteractiveShell.Interactable.__init__(self, key)

		def render(self, shell, offset, line, page = 0, index = -1, lines_before = -1, at_cursor = None):
			pass

		def hoverable(self):
			return True

		def placeholder(self):
			pass

		def observe_key(self, what, at_cursor = None):
			return False

class Separator(Shell.Interactable):
	def __init__(self, key = "separator", size = 1):
		Shell.Interactable.__init__(self, key)
		self.size = size

	def render(self, shell, offset, line):
		shell.write("\n" * self.size)

	def lines(self, shell):
		return self.size

class Notice(Shell.Interactable):
	def __init__(self, key, text = None):
		Shell.Interactable.__init__(self, key)
		self.text = text if text is not None else key

	def render(self, shell, offset, line):
		shell.write(str(self.text) + "\n")

	def lines(self, shell):
		return str(self.text).count("\n") + 1

class Entry(SelectiveShell.Selectable):
	def __init__(self, key, text = None, arrow = "> "):
		SelectiveShell.Selectable.__init__(self, key)
		self.text = text if text is not None else key
		self.arrow = arrow

	def get_arrow(self, at_cursor = None):
		return "" if at_cursor is None else \
			str(self.arrow) if at_cursor else " " * len(self.arrow)

	def render(self, shell, offset, line, page = 0, index = -1, lines_before = -1, at_cursor = None):
		shell.write(self.get_arrow(at_cursor) + str(self.text) + "\n")

	def placeholder(self):
		return str(self.text).partition("\n")[0]

	def lines(self, shell):
		return str(self.text).count("\n") + 1

class Switch(Entry):
	def __init__(self, key, text = None, checked = False, arrow = "> ", checked_arrow = "* ", hover_arrow = ">*"):
		Entry.__init__(self, key, text, arrow)
		self.checked_arrow = checked_arrow
		self.checked = checked
		self.hover_arrow = hover_arrow

	def get_arrow(self, at_cursor = None):
		return str(self.hover_arrow) if at_cursor and self.checked else \
			str(self.arrow) if at_cursor else str(self.checked_arrow) if self.checked else " " * len(self.arrow)

	def observe_key(self, what, at_cursor = None):
		if at_cursor and ord(what) in {10, 13}:
			self.checked = not self.checked
			return True
		return Entry.observe_key(self, what)

class Input(Entry):
	def __init__(self, key, hint = None, text = "", arrow = "> ", template = None, maximum_length = 40):
		Entry.__init__(self, key, text, arrow)
		self.hovered = False
		self.hint = hint
		self.template = template
		self.maximum_length = maximum_length

	def render(self, shell, offset, line, page = 0, index = -1, lines_before = -1, at_cursor = None):
		shell.write(self.get_arrow(at_cursor) + (str(self.hint) if self.hint is not None else "") +
			("" if self.hovered else "\x1b[2m") + (self.text if len(self.text) > 0 or self.hovered else \
				"..." if self.template is None else self.template) + ("\x1b[7m " if self.hovered else "") + "\x1b[0m\n")

	def read(self):
		return self.template if not self.hovered and len(self.text) == 0 and self.template is not None else self.text

	def observe_key(self, what, at_cursor = None):
		if at_cursor:
			if ord(what) in {10, 13}:
				self.hovered = not self.hovered
				return True
			if self.hovered:
				if what == "\x7f":
					if len(self.text) > 0:
						self.text = self.text[::-1][1:][::-1]
					else:
						self.hovered = False
				elif what.isprintable() and len(self.text + what) <= self.maximum_length:
					self.text += what
				return True
		return Entry.observe_key(self, what)

class Progress(Shell.Interactable):
	def __init__(self, key = "progress", progress = 0, weight = 49, text = None):
		Shell.Interactable.__init__(self, key)
		self.progress = progress
		self.weight = weight
		self.text = text

	def render(self, shell, offset, line):
		text = (str(self.text) if self.text is not None else str(int(self.progress * 100)) + "%").center(self.weight)
		size = int(self.weight * self.progress)
		shell.write("\x1b[7m" + text[:size] + "\x1b[2m" + text[size:self.weight] + "\x1b[0m\n")

	def seek(self, progress, text = None):
		self.progress = progress
		if text is not None:
			self.text = text

	def lines(self, shell):
		return (str(self.text).count("\n") if self.text is not None else 0) + 1

class Interrupt(InteractiveShell.Interactable):
	def __init__(self, key = "interrupt", occupied_page = True):
		InteractiveShell.Interactable.__init__(self, key)
		self.ocuppied_page = occupied_page

	def render(self, shell, offset, line, page = 0, index = -1, lines_before = -1):
		raise EOFError()

	def lines(self, shell):
		return shell.lines_per_page if self.ocuppied_page else 0

class Debugger(SelectiveShell.Selectable):
	def __init__(self, key = "debugger"):
		SelectiveShell.Selectable.__init__(self, key)

	def render(self, shell, offset, line, page = 0, index = -1, lines_before = -1, at_cursor = None):
		shell.write(f"Page {page}:{shell.global_buffer_offset}, offset {offset}/{shell.page_buffer_offset}, cursor {shell.page_cursor_offset}\n")

	def hoverable(self):
		return False

def select_prompt(prompt = None, *variants, fallback = None, what_not_which = False):
	if prompt is not None:
		print(prompt, end="")
	shell = SelectiveShell(infinite_scroll=True)
	for variant in variants:
		shell.interactables.append(Entry(variant))
	try:
		shell.loop()
		result = shell.which()
	except KeyboardInterrupt:
		result = fallback
	interactable = shell.get_interactable(result)
	try:
		print((prompt + " " if prompt is not None else "") + "\x1b[2m" + interactable.placeholder() + "\x1b[0m")
	except ValueError:
		pass
	return result if not what_not_which else interactable.key if interactable is not None else None


if __name__ == "__main__":
	shell = Shell()
	while True:
		try:
			key = shell.input(1)
		except KeyboardInterrupt:
			break
		print(ord(key), ": ", str(key.encode("unicode-escape"))[2:][::-1][1:][::-1].replace("\\\\", "\\"), sep="")



import sys
from os import listdir, environ, getenv, makedirs
from os.path import isfile, isdir, join, abspath, dirname
import platform
import subprocess
import re

from make_config import TOOLCHAIN_CONFIG
from shell import Notice, Progress, Shell
from utils import clear_directory, AttributeZipFile

def abi_to_arch(abi):
	abi_map = {
		"armeabi-v7a": "arm",
		"arm64-v8a": "arm64",
		"x86": "x86",
		"x86_64": "x86_64"
	}
	if abi in abi_map:
		return abi_map[abi]
	return None

def list_subdirectories(path, max_depth = 5, dirs = None):
	if dirs is None:
		dirs = []
	if not isdir(path):
		return dirs
	dirs.append(path)
	for filename in listdir(path):
		file = join(path, filename)
		if max_depth > 0 and isdir(file):
			list_subdirectories(file, dirs=dirs, max_depth=max_depth - 1)
	return dirs

def search_ndk_path(home_dir, contains_ndk = False):
	preferred_ndk_versions = [
		"android-ndk-r16b",
		"android-ndk-.*",
		"ndk-bundle"
	]
	possible_ndk_dirs = list_subdirectories(home_dir if contains_ndk else join(home_dir, "Android"))
	for ndk_dir_regex in preferred_ndk_versions:
		compiled_pattern = re.compile(ndk_dir_regex)
		for possible_ndk_dir in possible_ndk_dirs:
			if re.findall(compiled_pattern, possible_ndk_dir):
				return possible_ndk_dir

def get_ndk_path():
	path_from_config = TOOLCHAIN_CONFIG.get_value("ndkPath")
	if path_from_config is not None:
		path_from_config = TOOLCHAIN_CONFIG.get_absolute_path(path_from_config)
		if isdir(path_from_config):
			return path_from_config
	# Unix
	try:
		return search_ndk_path(environ["HOME"])
	except KeyError:
		pass
	# Windows
	return search_ndk_path(getenv("LOCALAPPDATA"))

def search_for_gcc_executable(ndk_dir):
	search_dir = join(ndk_dir, "bin")
	if isdir(search_dir):
		pattern = re.compile("[0-9A-Za-z]*-" + ("windows" if platform.system() == "Windows" else "linux") + "-android(eabi)*-g\\+\\+.*")
		for file in listdir(search_dir):
			if re.match(pattern, file):
				return abspath(join(search_dir, file))

def require_compiler_executable(arch, install_if_required = False):
	ndk_dir = TOOLCHAIN_CONFIG.get_path("toolchain/ndk/" + str(arch))
	file = search_for_gcc_executable(ndk_dir)
	if install_if_required:
		if install(arch=arch, reinstall=False) == -1:
			return None
		file = search_for_gcc_executable(ndk_dir)
		if file is None or not isfile(file):
			print("NDK installation for " + arch + " is broken, trying to re-install.")
			if install(arch=arch, reinstall=True) == -1:
				return None
			file = search_for_gcc_executable(ndk_dir)
			if file is None or not isfile(file):
				print("Reinstallation doesn't help, please, retry setup manually.")
				return None
		return file
	else:
		return file

def check_installed(arch):
	return isfile(TOOLCHAIN_CONFIG.get_path("toolchain/ndk/.installed-" + str(arch)))

def download(shell):
	from urllib import request
	archive_path = TOOLCHAIN_CONFIG.get_path("toolchain/temp/ndk.zip")
	makedirs(dirname(archive_path), exist_ok=True)

	if not isfile(archive_path):
		progress = Progress(text="C++ GCC Compiler (NDK)")
		shell.interactables.append(progress)
		shell.render()
		url = "https://dl.google.com/android/repository/android-ndk-r16b-" + ("windows" if platform.system() == "Windows" else "linux") + "-x86_64.zip"
		with request.urlopen(url) as response:
			with open(archive_path, "wb") as f:
				info = response.info()
				length = int(info["Content-Length"])
				downloaded = 0
				while True:
					buffer = response.read(8192)
					if not buffer:
						break
					downloaded += len(buffer)
					progress.seek(downloaded / length, f"Downloading ({(downloaded / 1048576):.1f}/{(length / 1048576):.1f}MiB)")
					shell.render()
					f.write(buffer)
		progress.seek(1, f"Downloaded {(length / 1024):.1f}MiB")
		shell.render()

	progress = Progress(text="Extracting NDK/GCC")
	shell.interactables.append(progress)
	shell.render()
	extract_path = TOOLCHAIN_CONFIG.get_path("toolchain/temp")
	with AttributeZipFile(archive_path, "r") as archive:
		archive.extractall(extract_path)
	progress.seek(1, "Extracted into toolchain/temp")
	shell.render()

	return search_ndk_path(extract_path, contains_ndk=True)

def install(arch = "arm", reinstall = False):
	if not reinstall and check_installed(arch):
		return 0
	else:
		shell = Shell()
		ndk_path = get_ndk_path()
		if ndk_path is None:
			if not reinstall:
				print("Not found valid NDK installation for ", arch, ".", sep="")
			try:
				if reinstall or input("Download android-ndk-r16b-x86_64? [N/y] ")[:1].lower() == "y":
					shell.enter()
					ndk_path = download(shell)
				else:
					print("Abort.")
					return -1
			except KeyboardInterrupt:
				print("Abort.")
				return -1
		else:
			shell.enter()

		progress = Progress(text="Installing")
		shell.interactables.append(progress)
		shell.render()
		result = subprocess.call([
			"python3" if platform.system() != "Windows" else "python",
			join(ndk_path, "build", "tools", "make_standalone_toolchain.py"),
			"--arch", str(arch),
			"--api", "19",
			"--install-dir", TOOLCHAIN_CONFIG.get_path("toolchain/ndk/" + str(arch)),
			"--force"
		])

		if result == 0:
			open(TOOLCHAIN_CONFIG.get_path("toolchain/ndk/.installed-" + str(arch)), "tw").close()
			progress.seek(0.9, "Removing temporary files")
			shell.render()
			clear_directory(TOOLCHAIN_CONFIG.get_path("toolchain/temp"))
			progress.seek(1, "Successfully installed")
			shell.render()
			shell.leave()
			return 0
		else:
			progress.seek(0.5, f"Installation failed with result {str(result)}")
			shell.render()
			notice = Notice("You are must install it manually from toolchain/temp/ndk.zip")
			shell.interactables.append(notice)
			shell.render()
			shell.leave()
			return result


if __name__ == "__main__":
	if "--help" in sys.argv:
		print("Usage: native/native-setup.py [arch]")
		exit(0)
	if len(sys.argv) >= 2:
		install(sys.argv[1])
	else:
		install()



import os
from os.path import join, isfile, isdir, relpath
import json

from utils import ensure_directory, clear_directory, ensure_file_dir
from make_config import MAKE_CONFIG

class BuildTargetType:
	def __init__(self, directory = None, list_property = None, **kw):
		self.directory = directory
		self.list_property = list_property

BUILD_TARGETS = {
	"resource_directory": BuildTargetType(directory=MAKE_CONFIG.get_value("target.resource_directory", "resources"), list_property="resources"),
	"gui": BuildTargetType(directory=MAKE_CONFIG.get_value("target.gui", "gui"), list_property="resources"),
	"minecraft_resource_pack": BuildTargetType(directory=MAKE_CONFIG.get_value("target.minecraft_resource_pack", "minecraft_packs/resource"), list_property="resources"),
	"minecraft_behavior_pack": BuildTargetType(directory=MAKE_CONFIG.get_value("target.minecraft_behavior_pack", "minecraft_packs/behavior"), list_property="resources"),

	"script_source": BuildTargetType(directory=MAKE_CONFIG.get_value("target.source", "source"), list_property="compile"),
	"script_library": BuildTargetType(directory=MAKE_CONFIG.get_value("target.library", "library"), list_property="compile"),
	"native": BuildTargetType(directory=MAKE_CONFIG.get_value("target.native", "native"), list_property="nativeDirs"),
	"java": BuildTargetType(directory=MAKE_CONFIG.get_value("target.java", "java"), list_property="javaDirs")
}

class ModStructure:
	def __init__(self, output_directory):
		self.directory = MAKE_CONFIG.get_path(output_directory)
		self.targets = {}
		self.build_config = None

	def cleanup_build_target(self, target_type_name):
		target_type = BUILD_TARGETS[target_type_name]
		self.targets[target_type_name] = []
		if target_type.directory == "":
			return
		directory = join(self.directory, target_type.directory)
		reldir = relpath(directory, self.directory)
		if reldir[:2] == "..":
			print(f"WARNING: Output target {target_type_name} is not relative to output, it will be not cleaned!")
			return
		clear_directory(directory)
		ensure_directory(directory)

	def new_build_target(self, target_type_name, name, **properties):
		target_type = BUILD_TARGETS[target_type_name]
		formatted_name = name.format("")
		if target_type_name in self.targets:
			targets_by_name = list(map(lambda x: x["name"], self.targets[target_type_name]))
			index = 0
			while formatted_name in targets_by_name:
				formatted_name = name.format(index)
				index += 1
		else:
			self.targets[target_type_name] = []

		target_path = join(self.directory, target_type.directory, formatted_name)
		self.targets[target_type_name].append({
			"name": formatted_name,
			"path": target_path,
			**properties
		})
		return target_path

	def get_all_targets(self, target_type, prop = None, values = ()):
		targets = []
		if target_type in self.targets:
			for target in self.targets[target_type]:
				if prop is None or prop in target and target[prop] in values:
					targets.append(target)
		return targets

	def get_target_directories(self, *names, filter_unchanged = False):
		return list(map(lambda name: BUILD_TARGETS[name].directory,
			filter(lambda name: name in self.targets \
				and len(self.targets[name]) > 0, names) if filter_unchanged else names
		))

	def create_build_config_list(self, list_name, default_overrides = None):
		result = []
		for target_name, target_type in BUILD_TARGETS.items():
			if target_type.list_property == list_name and target_name in self.targets:
				for target in self.targets[target_name]:
					if "exclude" not in target or not target["exclude"]:
						result.append({
							"path": target_type.directory + "/" + target["name"],
							**(target["declare"] if "declare" in target else {})
						})
					if default_overrides is not None and "declare_default" in target:
						for key, value in target["declare_default"].items():
							default_overrides[key] = value
		return result

	def read_or_create_build_config(self):
		build_config_file = join(self.directory, "build.config")
		if isfile(build_config_file):
			with open(build_config_file, "r", encoding="utf-8") as build_config:
				try:
					self.build_config = json.loads(build_config.read())
					return
				except json.JSONDecodeError as err:
					print("Something went wrong while reading cached build config:", err)
		self.build_config = {}

	def write_build_config(self):
		if self.build_config is None:
			return
		build_config_file = join(self.directory, "build.config")
		if isdir(build_config_file):
			clear_directory(build_config_file)
			os.remove(build_config_file)
		ensure_file_dir(build_config_file)
		with open(build_config_file, "w", encoding="utf-8") as build_config:
			build_config.write(json.dumps(self.build_config, indent=" " * 2))

	def setup_default_config(self):
		self.read_or_create_build_config()
		if "defaultConfig" not in self.build_config or not isinstance(self.build_config["defaultConfig"], dict):
			self.build_config["defaultConfig"] = {}
		default_config = self.build_config["defaultConfig"]
		default_config["readme"] = "this build config is generated automatically by mod development toolchain"
		default_config["api"] = MAKE_CONFIG.get_value("api", fallback="CoreEngine")
		optimization_level = MAKE_CONFIG.get_value("optimizationLevel")
		if optimization_level is not None:
			default_config["optimizationLevel"] = min(max(int(optimization_level), -1), 9)
		setup_script = MAKE_CONFIG.get_value("setupScript")
		if setup_script is not None:
			default_config["setupScript"] = setup_script
		default_config["buildType"] = "develop"
		if "buildDirs" not in self.build_config:
			self.build_config["buildDirs"] = []
		self.write_build_config()

	def update_build_config_list(self, list_name):
		self.setup_default_config()
		self.build_config[list_name] = self.create_build_config_list(list_name, default_overrides=self.build_config["defaultConfig"])
		self.write_build_config()


mod_structure = ModStructure("output")



import os
from os.path import join, basename, isfile, isdir, getmtime
import platform
import subprocess
import json
import hashlib
from zipfile import ZipFile

from utils import *
from component import which_installed, install_components
from make_config import MAKE_CONFIG, TOOLCHAIN_CONFIG
from mod_structure import mod_structure

def get_classpath_from_directories(directories):
	classpath = []
	for directory in directories:
		if isdir(directory):
			for file in os.listdir(directory):
				file = join(directory, file)
				if isfile(file):
					classpath.append(file)
	return classpath

def rebuild_library_cache(directory, library_files, cache_dir):
	directory_name = basename(directory)
	lib_cache_dir = join(cache_dir, "d8_lib_cache", directory_name)
	lib_cache_zip = join(cache_dir, "d8_lib_cache-" + directory_name + ".zip")

	print("Rebuilding library cache:", directory_name)
	clear_directory(lib_cache_dir)
	ensure_directory(lib_cache_dir)

	for lib_file in library_files:
		print("Extracting library classes:", basename(lib_file))
		with AttributeZipFile(lib_file, "r") as zip_ref:
			zip_ref.extractall(lib_cache_dir)
	print("Zipping")

	import shutil
	if isfile(lib_cache_zip):
		os.remove(lib_cache_zip)
	shutil.make_archive(lib_cache_zip[:-4], "zip", lib_cache_dir)
	return [lib_cache_zip]

def update_modified_classes(directories, cache_dir):
	cache_json = {}
	try:
		with open(join(cache_dir, "gradle_classes_cache.json")) as file:
			cache_json = json.load(file)
	except Exception:
		pass
	print()
	print("Recalculating class file hashes")
	modified_files = {}

	for directory in directories:
		directory_name = basename(directory)
		classes_dir = join(cache_dir, "classes", directory_name, "classes")
		if directory_name not in cache_json:
			cache_json[directory_name] = {}
		modified_timings = cache_json[directory_name]
		modified_files[directory_name] = {
			"class": [],
			"lib": []
		}
		modified_files_for_dir = modified_files[directory_name]["class"]
		for dirpath, dirnames, filenames in os.walk(classes_dir):
			for filename in filenames:
				if filename.endswith(".class"):
					file = str(join(dirpath, filename))
					modified_time = int(1000 * getmtime(file))
					hash_factory = hashlib.md5()
					with open(file, "rb") as fp:
						hash_factory.update(fp.read())
					hash = hash_factory.hexdigest()
					if file not in modified_timings or modified_timings[file] != hash:
						modified_files_for_dir.append(file)
					modified_timings[file] = hash

		with open(join(directory, "manifest"), "r") as file:
			manifest = json.load(file)

		was_libs_modified = False
		library_files = []
		for library_dir in manifest["library-dirs"]:
			for dirpath, dirnames, filenames in os.walk(join(directory, library_dir)):
				for filename in filenames:
					if filename.endswith(".jar"):
						file = join(dirpath, filename)
						library_files.append(file)
						modified_time = int(1000 * getmtime(file))
						key = "lib:" + file
						if key not in modified_timings or modified_timings[key] != modified_time:
							was_libs_modified = True
						modified_timings[key] = modified_time
		if was_libs_modified:
			modified_files[directory_name]["lib"] += rebuild_library_cache(directory, library_files, cache_dir)
	return modified_files, cache_json

def save_modified_classes_cache(cache_json, cache_dir):
	with open(join(cache_dir, "gradle_classes_cache.json"), "w") as file:
		file.write(json.dumps(cache_json))

def run_d8(directory_name, modified_files, cache_dir, debug_build = False):
	d8_libs = []
	classpath_dir = TOOLCHAIN_CONFIG.get_path("toolchain/classpath")
	if exists(classpath_dir):
		for dirpath, dirnames, filenames in os.walk(classpath_dir):
			for filename in filenames:
				if filename.endswith(".jar"):
					d8_libs += ["--lib", join(dirpath, filename)]

	dex_classes_dir = join(cache_dir, "d8", directory_name)
	jar_dir = join(cache_dir, "classes", directory_name, "libs", directory_name + "-all.jar")
	ensure_directory(dex_classes_dir)

	modified_classes = modified_files["class"]
	modified_libs = modified_files["lib"]

	print("Dexing libraries")
	result = subprocess.call([
		"java",
		"-cp", TOOLCHAIN_CONFIG.get_path("toolchain/bin/r8/r8.jar"),
		"com.android.tools.r8.D8"
	] + modified_libs + d8_libs + ([
		"--classpath", classpath_dir
	] if exists(classpath_dir) else []) + [
		"--min-api", "19",
		"--lib", jar_dir,
		"--debug" if debug_build else "--release",
		"--intermediate",
		"--output", dex_classes_dir
	])
	if result != 0:
		return result

	print("Dexing classes")
	index = 0
	max_span_size = 128
	while index < len(modified_classes):
		modified_classes_span = modified_classes[index:min(index + max_span_size, len(modified_classes))]
		result = subprocess.call([
			"java",
			"-cp", TOOLCHAIN_CONFIG.get_path("toolchain/bin/r8/r8.jar"),
			"com.android.tools.r8.D8"
		] + modified_classes_span + d8_libs + ([
			"--classpath", classpath_dir
		] if exists(classpath_dir) else []) + [
			"--min-api", "19",
			"--lib", jar_dir,
			"--debug" if debug_build else "--release",
			"--intermediate",
			"--file-per-class",
			"--output", dex_classes_dir
		])
		if result != 0:
			return result
		index += max_span_size
		print(f"Dexing classes: {min(index, len(modified_classes))}/{len(modified_classes)} completed")

	print("Compressing dex archives")
	dex_classes_dir = join(cache_dir, "d8", directory_name)
	dex_zip_file = join(cache_dir, "d8", directory_name + ".zip")
	with ZipFile(dex_zip_file, "w") as zip_ref:
		for dirpath, dirnames, filenames in os.walk(dex_classes_dir):
			for filename in filenames:
				if filename.endswith(".dex"):
					file = join(dirpath, filename)
					zip_ref.write(file, arcname=file[len(dex_classes_dir) + 1:])

	return result

def merge_compressed_dexes(directory_name, cache_dir, output_dex_dir, debug_build = False):
	dex_zip_file = join(cache_dir, "d8", directory_name + ".zip")
	print("Merging dex")
	return subprocess.call([
		"java",
		"-cp", TOOLCHAIN_CONFIG.get_path("toolchain/bin/r8/r8.jar"),
		"com.android.tools.r8.D8",
		dex_zip_file,
		"--min-api", "19",
		"--debug" if debug_build else "--release",
		"--intermediate",
		"--output", output_dex_dir
	])

def build_java_directories(directories, cache_dir, classpath, debug_build = False):
	ensure_directory(cache_dir)

	targets = setup_gradle_project(cache_dir, directories, classpath)
	gradle_executable = TOOLCHAIN_CONFIG.get_path("toolchain/bin/gradlew")
	if platform.system() == "Windows":
		gradle_executable += ".bat"
	result = subprocess.call([
		gradle_executable,
		"-p", cache_dir, "shadowJar"
	])
	if result != 0:
		print(f"Java compilation failed with code {result}")
		return result

	modified_files, cache_json = update_modified_classes(directories, cache_dir)
	save_modified_classes_cache(cache_json, cache_dir)
	for target in targets:
		directory_name = basename(target)
		if directory_name in modified_files and (len(modified_files[directory_name]["class"]) > 0 or len(modified_files[directory_name]["lib"]) > 0):
			print(f"\x1b[1m\x1b[92m\n* Running d8 for {directory_name}\x1b[0m\n")
			result = run_d8(directory_name, modified_files[directory_name], cache_dir, debug_build)
			if result != 0:
				print(f"Failed to dex {directory_name} with code {result}")
				return result
		else:
			print(f"* Directory {directory_name} is not changed")
		result = merge_compressed_dexes(directory_name, cache_dir, target, debug_build)
		if result != 0:
			print(f"Failed to merge {directory_name} with code {result}")
			return result

	print("\n\x1b[1m\x1b[92mJAVA BUILD COMPLETED\x1b[0m\n")
	return result

def build_list(working_dir):
	dirs = os.listdir(working_dir)
	if "order.txt" in dirs:
		order = open(join(working_dir, "order.txt"), "r", encoding="utf-8")
		dirs = order.read().splitlines()
	else:
		dirs = list(filter(lambda name: isdir(join(working_dir, name)), dirs))
	return dirs

def setup_gradle_project(cache_dir, directories, classpath):
	file = open(join(cache_dir, "settings.gradle"), "w", encoding="utf-8")
	file.writelines(["include \":%s\"\nproject(\":%s\").projectDir = file(\"%s\")\n"
				  % (basename(item), basename(item), item.replace("\\", "\\\\"))
				  for item in directories])
	file.close()

	targets = []
	for directory in directories:
		target_dir = mod_structure.new_build_target("java", basename(directory))
		clear_directory(target_dir)
		ensure_directory(target_dir)
		copy_file(join(directory, "manifest"), join(target_dir, "manifest"))
		targets.append(target_dir)

		with open(join(directory, "manifest"), "r", encoding="utf-8") as file:
			manifest = json.load(file)

		source_dirs = manifest["source-dirs"]
		library_dirs = manifest["library-dirs"]
		build_dir = join(cache_dir, "classes")
		dex_dir = target_dir
		ensure_directory(build_dir)
		ensure_directory(dex_dir)

		if MAKE_CONFIG.get_value("gradle.keepLibraries", True):
			for library_dir in library_dirs:
				src_dir = join(directory, library_dir)
				if isdir(src_dir):
					copy_directory(src_dir, join(dex_dir, library_dir), clear_dst=True)

		if MAKE_CONFIG.get_value("gradle.keepSources", False):
			for source_dir in source_dirs:
				src_dir = join(directory, source_dir)
				if isdir(src_dir):
					copy_directory(src_dir, join(dex_dir, source_dir), clear_dst=True)

		write_build_gradle(directory, classpath, build_dir, source_dirs, library_dirs)
	return targets

def write_build_gradle(directory, classpath, build_dir, source_dirs, library_dirs):
	with open(join(directory, "build.gradle"), "w", encoding="utf-8") as build_file:
		build_file.write("""plugins {
	id "com.github.johnrengelman.shadow" version "5.2.0"
	id "java"
}

dependencies {
	""" + ("""compile fileTree(\"""" + "\", \"".join([path.replace("\\", "\\\\")
												for path in library_dirs])
		+ """\") { include \"*.jar\" }""" if len(library_dirs) > 0 else "") + """
}

sourceSets {
	main {
		java {
			srcDirs = [\"""" + "\", \"".join([path.replace("\\", "\\\\")
								  for path in source_dirs]) + """\"]
			buildDir = \"""" + join(build_dir, "${project.name}").replace("\\", "\\\\") + """\"
		}
		resources {
			srcDirs = []
		}
		compileClasspath += files(\"""" + "\", \"".join([path.replace("\\", "\\\\")
												for path in classpath]) + """\")
	}
}
""")

def cleanup_gradle_scripts(directories):
	for path in directories:
		gradle_script = join(path, "build.gradle")
		if isfile(gradle_script):
			os.remove(gradle_script)

def compile_all_using_make_config(debug_build = False):
	import time
	start_time = time.time()

	overall_result = 0
	cache_dir = MAKE_CONFIG.get_build_path("gradle")
	ensure_directory(cache_dir)

	directories = []
	directory_names = []
	for directory in MAKE_CONFIG.get_filtered_list("compile", "type", ("java")):
		if "source" not in directory:
			print("Skipped invalid java directory json", directory)
			overall_result += 1
			continue

		for path in MAKE_CONFIG.get_paths(directory["source"]):
			if not isdir(path):
				print("Skipped non existing java directory path", directory["source"])
				overall_result += 1
				continue
			directories.append(path)

	if overall_result != 0:
		print("Java directories not found due to above error.")
		return overall_result

	if len(directories) > 0:
		if "java" not in which_installed():
			install_components(["java"])
			if "java" not in which_installed():
				from task import error
				error("Component java (r8) required for java compilation, nothing to do.")
		classpath_dir = TOOLCHAIN_CONFIG.get_path("toolchain/classpath")
		if not exists(classpath_dir):
			print("\x1b[93mNot found toolchain/classpath, in most cases build will be failed, please install it via tasks.\x1b[0m")
		classpath_directories = ([
			classpath_dir
		] if exists(classpath_dir) else []) + MAKE_CONFIG.get_value("gradle.classpath", [])
		overall_result = build_java_directories(directories, cache_dir, get_classpath_from_directories(classpath_directories), debug_build)
		if overall_result != 0:
			print(f"Nope, clearing compiled directories {directories}")
			for directory_name in directory_names:
				clear_directory(MAKE_CONFIG.get_path("output/" + directory_name))
	cleanup_gradle_scripts(directories)
	mod_structure.update_build_config_list("javaDirs")

	print(f"Completed java build in {int((time.time() - start_time) * 100) / 100}s with result {overall_result} - {'OK' if overall_result == 0 else 'ERROR'}")
	return overall_result


if __name__ == "__main__":
	compile_all_using_make_config(debug_build=True)



import os
from os.path import join, exists, isfile, isdir, basename, abspath
import json

from utils import clear_directory, ensure_not_whitespace
from make_config import MAKE_CONFIG, TOOLCHAIN_CONFIG
from workspace import CODE_WORKSPACE, CODE_SETTINGS

class ProjectManager:
	def __init__(self):
		self.projects = []
		self.templates = []
		locations = MAKE_CONFIG.get_value("projectLocations", [])
		for location in ["", *locations]:
			path = TOOLCHAIN_CONFIG.get_absolute_path(location)
			if not exists(path) or not isdir(path):
				print(f"Not found project location {location}!")
				continue
			for entry in ["", *os.listdir(path)]:
				make_path = join(path, entry, "make.json")
				if exists(make_path) and isfile(make_path):
					self.projects.append(join(location, entry))
				template_path = join(path, entry, "template.json")
				if exists(template_path) and isfile(template_path):
					self.templates.append(join(location, entry))

	def create_project(self, template, folder, name = None, author = None, version = None, description = None, clientOnly = False):
		location = TOOLCHAIN_CONFIG.get_path(folder)
		if exists(location):
			from task import error
			error(f"Folder '{folder}' already exists!")
		template_path = TOOLCHAIN_CONFIG.get_absolute_path(template)
		if not exists(template_path):
			from task import error
			error(f"Not found {template} template, nothing to do.")
		template_make_path = TOOLCHAIN_CONFIG.get_absolute_path(template + "/template.json")
		if not exists(template_make_path):
			from task import error
			error(f"Not found template.json in template {template}, nothing to do.")

		with open(template_make_path, "r", encoding="utf-8") as make_file:
			template_obj = json.loads(make_file.read())

		if not "info" in template_obj:
			template_obj["info"] = {}
		template_info = template_obj["info"]
		template_info["name"] = ensure_not_whitespace(name, ensure_not_whitespace(
			template_info["name"] if "name" in template_info else None, "Mod"
		))
		template_info["author"] = ensure_not_whitespace(author, ensure_not_whitespace(
			template_info["author"] if "author" in template_info else None, "ICMods"
		))
		template_info["version"] = ensure_not_whitespace(version, ensure_not_whitespace(
			template_info["version"] if "version" in template_info else None, "1.0"
		))
		template_info["description"] = description if description is not None else ensure_not_whitespace(
			template_info["description"] if "description" in template_info else None, "My brand new mod."
		)
		template_info["clientOnly"] = clientOnly if clientOnly is not None else \
			template_info["clientOnly"] if "clientOnly" in template_info else False

		os.makedirs(location)
		from package import setup_project
		setup_project(template_obj, template_path, location)

		make_path = join(location, "make.json")
		with open(make_path, "w", encoding="utf-8") as make_file:
			make_file.write(json.dumps(template_obj, indent="\t") + "\n")

		if CODE_WORKSPACE.available():
			location = CODE_WORKSPACE.get_toolchain_path(folder)
			if len(CODE_WORKSPACE.get_filtered_list("folders", "path", [location])) == 0:
				self.append_workspace_folder(folder, template_info["name"])

		if CODE_SETTINGS.available():
			exclude = CODE_SETTINGS.json["files.exclude"] if "files.exclude" in CODE_SETTINGS.json else {}
			if not folder.startswith("../"):
				exclude[folder] = True
				CODE_SETTINGS.json["files.exclude"] = exclude
				CODE_SETTINGS.save()

		self.projects.append(folder)
		return self.how_much() - 1

	def remove_project(self, index = None, folder = None):
		if len(self.projects) == 0:
			from task import error
			error("Not found any project to remove.")

		index, folder = self.get_folder(index, folder)

		if folder == MAKE_CONFIG.current_project:
			self.select_project_folder()

		if CODE_WORKSPACE.available():
			location = CODE_WORKSPACE.get_toolchain_path(folder)
			if len(CODE_WORKSPACE.get_filtered_list("folders", "path", [location])) > 0:
				folders = CODE_WORKSPACE.get_value("folders", [])
				for entry in folders:
					if isinstance(entry, dict) and "path" in entry and entry["path"] == location:
						folders.remove(entry)
				CODE_WORKSPACE.set_value("folders", folders)
				CODE_WORKSPACE.save()

		if CODE_SETTINGS.available():
			exclude = CODE_SETTINGS.json["files.exclude"] if "files.exclude" in CODE_SETTINGS.json else {}
			if folder in exclude:
				del exclude[folder]
				CODE_SETTINGS.json["files.exclude"] = exclude
				CODE_SETTINGS.save()

		clear_directory(TOOLCHAIN_CONFIG.get_absolute_path(folder))
		del self.projects[index]

	def append_workspace_folder(self, folder, name = "Mod"):
		if CODE_WORKSPACE.available():
			folders = CODE_WORKSPACE.get_value("folders", [])
			if len(folders) == 0:
				folders.append({
					"path": CODE_WORKSPACE.get_toolchain_path(),
					"name": "Inner Core Mod Toolchain"
				})
			folders.append({
				"path": CODE_WORKSPACE.get_toolchain_path(folder),
				"name": str(name)
			})
			CODE_WORKSPACE.set_value("folders", folders)
			CODE_WORKSPACE.save()

	def select_project_folder(self, folder = None):
		if MAKE_CONFIG.current_project == folder:
			return

		if folder is None:
			TOOLCHAIN_CONFIG.remove_value("currentProject")
		else:
			TOOLCHAIN_CONFIG.set_value("currentProject", folder)
		TOOLCHAIN_CONFIG.save()

		MAKE_CONFIG.__init__(TOOLCHAIN_CONFIG.filename)
		if folder is None:
			TOOLCHAIN_CONFIG.__init__(TOOLCHAIN_CONFIG.filename)
		else:
			TOOLCHAIN_CONFIG.__init__(MAKE_CONFIG.prototype.filename, MAKE_CONFIG.prototype)
		MAKE_CONFIG.prototype = TOOLCHAIN_CONFIG if folder is not None else None

	def select_project(self, index = None, folder = None):
		index, folder = self.get_folder(index, folder)

		if folder is not None and CODE_WORKSPACE.available():
			location = CODE_WORKSPACE.get_toolchain_path(folder)
			if len(CODE_WORKSPACE.get_filtered_list("folders", "path", [location])) == 0:
				make_path = TOOLCHAIN_CONFIG.get_absolute_path(folder + "/make.json")
				if not exists(make_path):
					from task import error
					error(f"Not found make.json in project {folder}, nothing to do.")
				with open(make_path, "r", encoding="utf-8") as make_file:
					make_obj = json.loads(make_file.read())
				self.append_workspace_folder(folder, make_obj["info"]["name"] if "info" in make_obj and "name" in make_obj["info"] else basename(folder))

		if folder is not None and CODE_SETTINGS.available():
			exclude = CODE_SETTINGS.json["files.exclude"] if "files.exclude" in CODE_SETTINGS.json else {}
			if MAKE_CONFIG.current_project is not None and not MAKE_CONFIG.current_project.startswith("../") and not exists(abspath(MAKE_CONFIG.current_project)):
				exclude[MAKE_CONFIG.current_project] = True
			if not folder.startswith("../"):
				exclude[folder] = False
			CODE_SETTINGS.json["files.exclude"] = exclude
			CODE_SETTINGS.save()

		self.select_project_folder(folder)
		print(f"Project '{folder}' selected.")

	def get_folder(self, index = None, folder = None):
		if index == None:
			if folder == None:
				raise ValueError("Folder index must be specified!")
			else:
				index = next((i for i, x in enumerate(self.projects)
					if x.lower() == folder.lower()
				), -1)

		folder = self.projects[index]
		return index, folder

	def how_much(self):
		return len(self.projects)

	def require_selection(self, prompt = None, prompt_when_single = None, dont_want_anymore = None):
		from package import select_project
		if self.how_much() == 1:
			itwillbe = self.projects[0]
			if prompt_when_single is None:
				return itwillbe
			else:
				if input(prompt_when_single.format(itwillbe) + " [Y/n] ")[:1].lower() == "n":
					return print("Abort.")
				return itwillbe
		if dont_want_anymore is not None:
			who = self.projects.copy()
			who.append(dont_want_anymore)
		else:
			who = self.projects
		raw = select_project(who, prompt, MAKE_CONFIG.current_project)
		return (raw if raw != dont_want_anymore else None)


PROJECT_MANAGER = ProjectManager()



import os
from os.path import join, exists, basename, isfile, normpath, splitext, relpath
import glob
import json
import re
import subprocess

from utils import move_file, copy_file
from make_config import MAKE_CONFIG, TOOLCHAIN_CONFIG
from hash_storage import build_storage

params_list = {
	"allowJs": False,
	"allowUnusedLabels": False,
	"alwaysStrict":	 False,
	"assumeChangesOnlyAffectDirectDependencies": False,
	"checkJs": False,
	"declaration": False,
	"diagnostics": False,
	"disableSizeLimit":	False,
	"downlevelIteration": False,
	"emitDeclarationOnly": False,
	"experimentalDecorators": False,
	"extendedDiagnostics": False,
	"importHelpers": False,
	"listEmittedFiles":	False,
	"listFiles": False,
	"locale": "en",
	"noEmit": False,
	"noEmitHelpers": False,
	"noEmitOnError": False,
	"noErrorTruncation": False,
	"noFallthroughCasesInSwitch": False,
	"noLib": False,
	"noResolve": False,
	"noStrictGenericChecks": False,
	"noUnusedLocals": False,
	"noUnusedParameters": False,
	"preserveConstEnums": False,
	"preserveSymlinks":	False,
	"pretty": True,
	"removeComments": False,
	"showConfig": False,
	"skipLibCheck": False,
	"sourceMap": False,
	"strict": False,
	"tsBuildInfoFile": ".tsbuildinfo"
}

temp_directory = MAKE_CONFIG.get_build_path("sources")

class Includes:
	def __init__(self, directory, includes_file, debug_build = False):
		self.file = join(directory, includes_file)
		self.includes_file = includes_file
		self.directory = directory
		self.debug_build = debug_build

		self.include = []
		self.exclude = []
		self.params = {}

	def read(self):
		with open(self.file, encoding="utf-8") as includes:
			for line in includes:
				line = line.strip()
				self.decode_line(line)

	def decode_line(self, line):
		if line.startswith("#"): # comment or parameter
			pair = [item.strip() for item in line[1:].strip().split(":")]
			key = pair[0]

			if key in params_list:
				default = params_list[key]

				if len(pair) > 1:
					v = pair[1].lower()
					self.params[key] = v == "true" if v in ["true", "false"] else v
				else:
					self.params[key] = True if isinstance(default, bool) else default

		elif len(line) < 1 or line.startswith("//"):
			return

		elif line.startswith("!"):
			line = line[1:].strip()
			search_path = (join(self.directory, line[:-2], ".") + "/**/*") \
				if line.endswith("/.") else join(self.directory, line)
			for file in glob.glob(search_path, recursive=True):
				file = normpath(file)
				if file not in self.include:
					self.exclude.append(relpath(file, self.directory).replace("\\", "/"))

		else:
			search_path = re.sub(r"\.$", "**/*", line) if line.endswith("/.") else line
			self.include.append(search_path.replace("\\", "/"))

	def create(self):
		with open(self.file, "w") as includes:
			params = ["# " + key + ((": " + str(self.params[key])).lower() if not self.params[key] else "") + "\n" for key in self.params]
			files = [file + "\n" for file in self.include]

			includes.writelines(params)
			includes.write("\n")
			includes.writelines(files)

	@staticmethod
	def create_from_directory(directory, includes_file, debug_build = False):
		includes = Includes(directory, includes_file, debug_build)
		includes.files = [normpath(relpath(file, directory))
			for file in glob.glob(f"{directory}/**/*", recursive=True)]
		includes.params = {}
		includes.create()

		return includes

	@staticmethod
	def create_from_tsconfig(directory, includes_file, debug_build = False):
		with open(join(directory, "tsconfig.json")) as tsconfig:
			config = json.load(tsconfig)

			params = config["compilerOptions"]
			include = config["include"]
			exclude = config["exclude"]

			if "target" in params:
				del params["target"]
			if "lib" in params:
				del params["lib"]
			if "outFile" in params:
				del params["outFile"]

		includes = Includes(directory, includes_file, debug_build)
		includes.include = include
		includes.exclude = exclude
		includes.params = params
		includes.create()

		return includes

	@staticmethod
	def invalidate(directory, includes_file, debug_build = False):
		if not isfile(join(directory, includes_file)):
			tsconfig_path = join(directory, "tsconfig.json")
			if isfile(tsconfig_path):
				includes = Includes.create_from_tsconfig(directory, includes_file, debug_build)
			else:
				includes = Includes.create_from_directory(directory, includes_file, debug_build)
		else:
			includes = Includes(directory, includes_file, debug_build)
		includes.read()

		return includes

	def build(self, target_path, language = "typescript"):
		temp_path = join(temp_directory, basename(target_path))

		result = 0
		if language == "typescript":
			self.create_tsconfig(temp_path)
		if build_storage.is_path_changed(self.directory):
			import datetime

			print(f"Building {basename(target_path)} from {self.includes_file}")

			start_time = datetime.datetime.now()
			result = self.build_source(temp_path, language)
			end_time = datetime.datetime.now()
			diff = end_time - start_time

			print(f"Completed {basename(target_path)} build in {round(diff.total_seconds(), 2)}s with result {result} - {'OK' if result == 0 else 'ERROR'}")
			if result != 0:
				return result
			build_storage.save()
		else:
			print(f"* Build target {basename(target_path)} is not changed")
		copy_file(temp_path, target_path)

		return result

	def get_tsconfig(self):
		return join(self.directory, "tsconfig.json")

	def create_tsconfig(self, temp_path):
		declarations = []
		if exists(TOOLCHAIN_CONFIG.get_path("toolchain/declarations")):
			declarations.extend(glob.glob(
				TOOLCHAIN_CONFIG.get_path("toolchain/declarations/**/*.d.ts"),
				recursive=True
			))
		declarations.extend(glob.glob(
			MAKE_CONFIG.get_build_path("declarations/**/*.d.ts"),
			recursive=True
		))

		current_name = splitext(basename(temp_path))[0]
		for declaration in declarations:
			if declaration.endswith(f"{current_name}.d.ts"):
				declarations.remove(declaration)

		if self.debug_build:
			for excluded in MAKE_CONFIG.get_value("debugIncludesExclude", []):
				if exists(str(excluded).lstrip("/").partition("/")[0]):
					for declaration in glob.glob(excluded, recursive=True):
						if declaration in declarations:
							declarations.remove(declaration)
				else:
					for declaration in glob.glob(TOOLCHAIN_CONFIG.get_path(excluded), recursive=True):
						if declaration in declarations:
							declarations.remove(declaration)

		template = {
			"compilerOptions": {
				"target": "ES5",
				"lib": ["ESNext"],
				"outFile": temp_path,
				"experimentalDecorators": True,
				"declaration": True,
				"downlevelIteration": True,
				"allowJs": True
			},
			"compileOnSave": False,
			"exclude": [
				"**/node_modules/*",
				"dom",
				"webpack"
			] + self.exclude,
			"include": self.include,
		}

		if len(declarations) > 0:
			template["files"] = declarations

		for key, value in self.params.items():
			template["compilerOptions"][key] = value

		with open(self.get_tsconfig(), "w") as tsconfig:
			json.dump(template, tsconfig, indent="\t")
			tsconfig.write("\n")

	def build_source(self, temp_path, language):
		if language.lower() == "typescript":
			command = [
				"tsc",
				"-p", self.get_tsconfig(),
				"--noEmitOnError"
			]
			if self.debug_build:
				command.append("--skipLibCheck")
				command.append("--noResolve")
			result = subprocess.call(command)
		else:
			result = 0

		declaration_path = f"{splitext(temp_path)[0]}.d.ts"
		if isfile(declaration_path):
			move_file(declaration_path, MAKE_CONFIG.get_build_path(
				"declarations/" + basename(declaration_path)
			))

		return result



import os
from os.path import join, exists, isfile, isdir
import platform
import shutil
import sys
import subprocess
import urllib.request as request
from urllib.error import URLError
from zipfile import ZipFile, ZipInfo

class AttributeZipFile(ZipFile):
	if sys.version_info < (3, 6):
		def extract(self, member, path = None, pwd = None):
			if not isinstance(member, ZipInfo):
				member = self.getinfo(member)

			if path is None:
				path = os.getcwd()

			targetpath = self._extract_member(member, path, pwd)

			attr = member.external_attr >> 16
			os.chmod(targetpath, attr)
			return targetpath

	else:
		def _extract_member(self, member, targetpath, pwd):
			if not isinstance(member, ZipInfo):
				member = self.getinfo(member)

			targetpath = super()._extract_member(member, targetpath, pwd)

			attr = member.external_attr >> 16
			if attr != 0:
				os.chmod(targetpath, attr)
			return targetpath

def download_and_extract_toolchain(directory):
	os.makedirs(directory, exist_ok=True)
	archive = join(directory, "toolchain.zip")

	readable_name = "current directory" if directory in (".", "") else "'" + directory + "'"
	if exists(join(directory, "toolchain")):
		print("Inner Core Mod Toolchain already installed in " + readable_name + ".")
		print("Newly installed files will be merged with your installation.")
		print("It's handly to restore necessary removed script or template.")
		try:
			if input("Do you want to download it again? [N/y] ")[:1].lower() != "y":
				return print("Abort.")
		except KeyboardInterrupt:
			return print("Abort.")

	if not exists(archive):
		url = "https://codeload.github.com/zheka2304/innercore-mod-toolchain/zip/deploy"
		print("Downloading Inner Core Mod Toolchain: " + url)
		try:
			request.urlretrieve(url, archive)
		except URLError:
			print("Check your network connection!")
			exit(1)
		except BaseException as err:
			print(err)
			print("Inner Core Mod Toolchain installation not completed due to above error.")
			exit(2)
	else:
		print("'toolchain.zip' already exists in " + readable_name + ".")

	print("Extracting into " + readable_name)
	with AttributeZipFile(archive, "r") as zip_ref:
		zip_ref.extractall(directory)

	commit = "unknown"
	try:
		if isdir(join(directory, "toolchain-mod/toolchain")):
			dirname = "toolchain-mod"
			index = 0
			while exists(join(directory, dirname)):
				index += 1
				dirname = "toolchain-mod-" + str(index)
			shutil.move(join(directory, "toolchain-mod"), join(directory, dirname))
		shutil.copytree(join(directory, "innercore-mod-toolchain-deploy"), directory, dirs_exist_ok=True)
		if isfile(join(directory, "toolchain/toolchain/bin/.commit")):
			with open(join(directory, "toolchain/toolchain/bin/.commit")) as file:
				commit = file.read()
		shutil.rmtree(join(directory, "innercore-mod-toolchain-deploy"))
	except BaseException as err:
		print(err)
		print("Inner Core Mod Toolchain installation not completed due to above error.")
		exit(3)
	finally:
		if not exists(join(directory, "toolchain")):
			print("Inner Core Mod Toolchain extracted 'innercore-mod-toolchain-deploy' folder not found.")
			print("Retry operation or extract 'toolchain.zip' manually.")
			exit(4)
		else:
			os.remove(archive)

	print("Installed into " + readable_name + " under " + commit.strip()[:7] + " revision.")

def print_placeholder(which):
	layer = 0
	while layer < len(which):
		for symbol in which[layer]:
			print("\x1b[", "" if symbol == 0 or symbol == 7 else "7m" if not isinstance(symbol, int) else "7m\x1b[" if symbol == 2 else "48;5;", symbol, "m  " if isinstance(symbol, int) else "", sep="", end="\x1b[0m")
		print()
		layer += 1


if "--help" in sys.argv:
	print("Usage: toolchain-setup.py <options> [directory]")
	print(" " * 2 + "--no-startup: Skip startup stage for configuring author")
	print(" " * 4 + "identity and installing additional components.")
	print(" " * 2 + "--import <folder> [destination]:")
	print(" " * 4 + "Run import after installation completion in specified")
	print(" " * 4 + "location, toolchain destination used by default.")
	print(" " * 2 + "--foreign: Skip startup stage, install components instead.")
	print("Download toolchain deploy branch, run startup script to complete")
	print("components installation and setup defaults.")
	exit(0)

location = sys.argv[len(sys.argv) - 1] if len(sys.argv) > 1 and not sys.argv[len(sys.argv) - 1].startswith("--") else "."
download_and_extract_toolchain(location)

if platform.system() != "Windows":
	print()
	print_placeholder([
		[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 124, 0, 0, 0, 0, 0, 0, 0, 0, 0],
		[0, 0, 0, 0, 0, 0, 0, 0, 196, 160, 196, 196, 196, 196, 0, 0, 0, 0, 0, 0, 196, 160, 0, 0, 0, 0, 0, 0, 0],
		[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 196, 196, 196, 196, 160, 0, 0, 0, 0, 196, 0, 0, 0, 0, 0, 0, 0],
		[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 196, 196, 196, 196, 0, 0, 0, 196, 196, 0, 0, 0, 0, 0, 0],
		[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 160, 196, 196, 160, 196, 160, 160, 0, 160, 0, 0, 196, 196, 0, 0, 0, 0, 0],
		[0, 0, 0, 0, 0, 160, 196, 196, 196, 196, 0, 0, 160, 196, 196, 196, 0, 220, 0, 160, 196, 0, 196, 196, 0, 0, 0, 0, 0],
		[0, 0, 0, 160, 196, 196, 196, 196, 196, 0, 0, 0, 226, 220, 196, 196, 196, 214, 220, 0, 196, 160, 196, 196, 196, 0, 0, 0, 0],
		[0, 0, 196, 196, 196, 160, 0, 196, 196, 196, 202, 208, 220, 226, 226, 202, 202, 196, 226, 226, 196, 196, 196, 196, 196, 0, 0, 0, 160],
		[0, 196, 196, 0, 0, 160, 196, 196, 196, 196, 196, 208, 214, 214, 214, 220, 2, 214, 220, 226, 196, 196, 0, 160, 196, 0, 0, 0, 196],
		[160, 160, 0, 0, 160, 196, 160, 172, 202, 214, 214, 2, 2, 7, 7, 7, 2, 7, 2, 214, 196, 196, 0, 0, 196, 0, 0, 0, 196],
		[0, 0, 0, 0, 0, 0, 0, 0, 226, 226, 2, 7, 7, 7, 7, 7, 7, 7, 7, 2, 214, 196, 0, 0, 0, 160, 0, 160, 196],
		[0, 0, 0, 160, 196, 0, 0, 166, 208, 226, 2, 7, 7, 7, 7, 7, 7, 7, 7, 2, 220, 208, 226, 0, 196, 196, 0, 196, 196],
		[0, 0, 0, 196, 196, 0, 196, 196, 214, 226, 7, 7, " I", "nn", "er", " C", "or", "e ", 7, 7, 214, 226, 226, 124, 196, 196, 0, 196, 160],
		[0, 0, 196, 196, 196, 160, 196, 196, 214, 2, 7, 7, " M", "od", 7, 7, 7, 7, 7, 7, 2, 226, 208, 196, 196, 0, 196, 196, 0],
		[0, 0, 196, 196, 0, 196, 196, 214, 220, 2, 7, 7, " T", "oo", "lc", "ha", "in", 7, 7, 7, 2, 214, 196, 196, 196, 196, 196, 196, 0],
		[0, 160, 196, 0, 196, 196, 124, 226, 220, 2, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 2, 214, 196, 202, 0, 196, 196, 0, 0]
	])
print()

if "--foreign" in sys.argv:
	subprocess.run([
		"python" if platform.system() == "Windows" else "python3",
		"component.py"
	], cwd=join(location, "toolchain\\toolchain\\python") if platform.system() == "Windows" else join(location, "toolchain/toolchain/python"))
elif not "--no-startup" in sys.argv:
	subprocess.run([
		"python" if platform.system() == "Windows" else "python3",
		"component.py", "--startup"
	], cwd=join(location, "toolchain\\toolchain\\python") if platform.system() == "Windows" else join(location, "toolchain/toolchain/python"))
if "--import" in sys.argv:
	where = sys.argv.index("--import")
	if len(sys.argv) < where + 1 or sys.argv[where + 1].startswith("--"):
		print("Not found import path, nothing will happened.")
		exit(5)
	folder = sys.argv[where + 1]
	subprocess.run([
		"python" if platform.system() == "Windows" else "python3",
		"import.py", folder
	], cwd=join(location, "toolchain\\toolchain\\python") if platform.system() == "Windows" else join(location, "toolchain/toolchain/python"))
